# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from typing import Any
from io import BytesIO
from collections.abc import Sequence, Callable
from ctypes import (
    get_last_error, WinError, FormatError,  # @UnresolvedImport
    sizeof, byref, cast, memset, memmove, create_string_buffer, c_char, c_void_p,
)
from ctypes.wintypes import HBITMAP

from xpra.os_util import gi_import
from xpra.platform.win32.common import (
    GetDC, ReleaseDC,
    WNDCLASSEX, GetLastError,
    WNDPROC, LPCWSTR, LPWSTR, LPCSTR, DWORD,
    BITMAPINFOHEADER, PBITMAPV5HEADER, BITMAPINFO,
    DefWindowProcW,
    GetModuleHandleA, RegisterClassExW, UnregisterClassW,
    CreateWindowExW, DestroyWindow,
    OpenClipboard, EmptyClipboard, CloseClipboard, GetClipboardData,
    GlobalLock, GlobalUnlock, GlobalAlloc, GlobalFree, GlobalSize,
    WideCharToMultiByte, MultiByteToWideChar,
    AddClipboardFormatListener, RemoveClipboardFormatListener,
    SetClipboardData, EnumClipboardFormats, GetClipboardFormatNameA, GetClipboardOwner,
    RegisterClipboardFormatA,
    GetWindowThreadProcessId, QueryFullProcessImageNameA, OpenProcess, CloseHandle,
    CreateDIBitmap,
)
from xpra.platform.win32 import win32con
from xpra.clipboard.timeout import ClipboardTimeoutHelper
from xpra.clipboard.core import (
    ClipboardProxyCore, log, _filter_targets, TEXT_TARGETS, MAX_CLIPBOARD_PACKET_SIZE, ClipboardCallback,
)
from xpra.common import roundup, noop
from xpra.util.str_fn import csv, Ellipsizer, bytestostr
from xpra.util.env import envint, envbool
from xpra.platform.win32.constants import PROCESS_QUERY_INFORMATION

GLib = gi_import("GLib")

CP_UTF8 = 65001
MB_ERR_INVALID_CHARS = 0x00000008
GMEM_MOVEABLE = 0x0002

WM_CLIPBOARDUPDATE = 0x031D

CLIPBOARD_EVENTS = {
    win32con.WM_CLEAR: "CLEAR",
    win32con.WM_CUT: "CUT",
    win32con.WM_COPY: "COPY",
    win32con.WM_PASTE: "PASTE",
    win32con.WM_ASKCBFORMATNAME: "ASKCBFORMATNAME",
    win32con.WM_CHANGECBCHAIN: "CHANGECBCHAIN",
    WM_CLIPBOARDUPDATE: "CLIPBOARDUPDATE",
    win32con.WM_DESTROYCLIPBOARD: "DESTROYCLIPBOARD",
    win32con.WM_DRAWCLIPBOARD: "DRAWCLIPBOARD",
    win32con.WM_HSCROLLCLIPBOARD: "HSCROLLCLIPBOARD",
    win32con.WM_PAINTCLIPBOARD: "PAINTCLIPBOARD",
    win32con.WM_RENDERALLFORMATS: "RENDERALLFORMATS",
    win32con.WM_RENDERFORMAT: "RENDERFORMAT",
    win32con.WM_SIZECLIPBOARD: "SIZECLIPBOARD",
    win32con.WM_VSCROLLCLIPBOARD: "WM_VSCROLLCLIPBOARD",
}

CLIPBOARD_FORMATS = {
    win32con.CF_BITMAP: "CF_BITMAP",
    win32con.CF_DIB: "CF_DIB",
    win32con.CF_DIBV5: "CF_DIBV5",
    win32con.CF_ENHMETAFILE: "CF_ENHMETAFILE",
    win32con.CF_METAFILEPICT: "CF_METAFILEPICT",
    win32con.CF_OEMTEXT: "CF_OEMTEXT",
    win32con.CF_TEXT: "CF_TEXT",
    win32con.CF_UNICODETEXT: "CF_UNICODETEXT",
}

BI_RGB = 0x0000
BI_RLE8 = 0x0001
BI_RLE4 = 0x0002
BI_BITFIELDS = 0x0003
BI_JPEG = 0x0004
BI_PNG = 0x0005
BI_CMYK = 0x000B
BI_CMYKRLE8 = 0x000C
BI_CMYKRLE4 = 0x000D
BI_FORMATS = {
    BI_RGB: "RGB",
    BI_RLE8: "RLE8",
    BI_RLE4: "RLE4",
    BI_BITFIELDS: "BITFIELDS",
    BI_JPEG: "JPEG",
    BI_PNG: "PNG",
    BI_CMYK: "CMYK",
    BI_CMYKRLE8: "CMYKRLE8",
    BI_CMYKRLE4: "CMYKRLE4",
}

LCS_CALIBRATED_RGB = 0x00000000
LCS_sRGB = 0x73524742
LCS_WINDOWS_COLOR_SPACE = 0x57696E20
PROFILE_LINKED = 3
PROFILE_EMBEDDED = 4
COLOR_PROFILES = {
    LCS_CALIBRATED_RGB: "CALIBRATED_RGB",
    LCS_sRGB: "sRGB",
    LCS_WINDOWS_COLOR_SPACE: "WINDOWS",
    PROFILE_LINKED: "PROFILE_LINKED",
    PROFILE_EMBEDDED: "PROFILE_EMBEDDED",
}

RETRY = envint("XPRA_CLIPBOARD_RETRY", 5)
DELAY = envint("XPRA_CLIPBOARD_INITIAL_DELAY", 10)
CONVERT_LINE_ENDINGS = envbool("XPRA_CONVERT_LINE_ENDINGS", True)
log("win32 clipboard: RETRY=%i, DELAY=%i, CONVERT_LINE_ENDINGS=%s",
    RETRY, DELAY, CONVERT_LINE_ENDINGS)


# can be used to blocklist problematic clipboard peers:
# ie: VBoxTray.exe


def get_clients(envvar, default="") -> Sequence[str]:
    return tuple(x for x in os.environ.get("XPRA_%s_CLIPBOARD_CLIENTS" % envvar, default).split(",") if x)


BLOCKLISTED_CLIPBOARD_CLIENTS = get_clients("BLOCKLISTED")
SYNCDELAY_CLIPBOARD_CLIENTS = get_clients("NOSYNC", "VBoxTray.exe")
log("BLOCKLISTED_CLIPBOARD_CLIENTS=%s", BLOCKLISTED_CLIPBOARD_CLIENTS)
log("SYNCDELAY_CLIPBOARD_CLIENTS=%s", SYNCDELAY_CLIPBOARD_CLIENTS)
COMPRESSED_IMAGES = envbool("XPRA_CLIPBOARD_COMPRESSED_IMAGES", True)

CLIPBOARD_WINDOW_CLASS_NAME = "XpraWin32Clipboard"


def is_blocklisted(owner_info) -> bool:
    return any(owner_info.find(x) >= 0 for x in BLOCKLISTED_CLIPBOARD_CLIENTS)


def is_syncdelay(owner_info) -> bool:
    return any(owner_info.find(x) >= 0 for x in SYNCDELAY_CLIPBOARD_CLIENTS)


# initialize the window we will use
# for communicating with the OS clipboard API:

def get_owner_info(owner, our_window) -> str:
    if not owner:
        return "unknown"
    if owner == our_window:
        return "our window (hwnd=%#x)" % our_window
    pid = DWORD(0)
    GetWindowThreadProcessId(owner, byref(pid))
    if not pid:
        return "unknown (hwnd=%#x)" % owner
    # log("get_owner_info(%#x) pid=%s", owner, pid.value)
    proc_handle = OpenProcess(PROCESS_QUERY_INFORMATION, False, pid)
    if not proc_handle:
        return "pid %i (hwnd=%#x)" % (pid.value, owner)
    try:
        size = DWORD(256)
        process_name = create_string_buffer(size.value + 1)
        if not QueryFullProcessImageNameA(proc_handle, 0, process_name, byref(size)):
            return "pid %i" % pid.value
        return "'%s' with pid %s (hwnd=%#x)" % (bytestostr(process_name.value), pid.value, owner)
    finally:
        CloseHandle(proc_handle)


def with_clipboard_lock(window, success_callback: Callable, failure_callback: Callable, retries=RETRY, delay=DELAY):
    log("with_clipboard_lock%s", (window, success_callback, failure_callback, retries, delay))
    r = OpenClipboard(window)
    if r:
        log("OpenClipboard(%#x)=%s", window, r)
        try:
            r = success_callback()
            log("%s()=%s", success_callback, r)
            if r:
                return
        finally:
            r = CloseClipboard()
            log("CloseClipboard()=%s", r)
    e = WinError(GetLastError())
    owner = GetClipboardOwner()
    log("OpenClipboard(%#x)=%s, current owner: %s", window, e, get_owner_info(owner, window))
    if retries <= 0:
        failure_callback("OpenClipboard: too many failed attempts, giving up")
        return
    # try again later:
    GLib.timeout_add(delay, with_clipboard_lock,
                     window, success_callback, failure_callback, retries - 1, delay + 5)


def format_name(fmt: int) -> str:
    name = CLIPBOARD_FORMATS.get(fmt)
    if name:
        return name
    ulen = 128
    buf = LPCSTR(b" " * ulen)
    r = GetClipboardFormatNameA(fmt, buf, ulen - 1)
    value = buf.value
    if r == 0 or not value:
        return str(fmt)
    return value[:r].decode("latin1")


def format_names(fmts) -> Sequence[str]:
    return tuple(format_name(x) for x in fmts)


def get_clipboard_formats() -> Sequence[int]:
    formats = []
    fmt = 0
    while True:
        fmt = EnumClipboardFormats(fmt)
        if fmt:
            formats.append(fmt)
        else:
            break
    log("get_clipboard formats()=%s", csv(format_name(x) for x in formats))
    return formats


def w_to_utf8(data):
    wstr = cast(data, LPCWSTR)
    ulen = WideCharToMultiByte(CP_UTF8, 0, wstr, -1, None, 0, None, None)
    if ulen > MAX_CLIPBOARD_PACKET_SIZE:
        raise ValueError("unicode data is too large: %i bytes" % ulen)
    buftype = c_char * ulen
    buf = buftype()
    length = WideCharToMultiByte(CP_UTF8, 0, wstr, -1, byref(buf), ulen, None, None)
    if length == 0:
        raise ValueError("failed to convert to UTF8: %s" % FormatError(get_last_error()))
    if buf.raw[length - 1:length] == b"\0":
        b = buf.raw[:length - 1]
    else:
        b = buf.raw[:length]
    log("got %i UTF8 bytes: %s", len(b), Ellipsizer(b))
    if CONVERT_LINE_ENDINGS:
        return b.decode("utf8").replace("\r\n", "\n").encode("utf8")
    return b


def rgb_to_bitmap(img_data) -> HBITMAP:
    from PIL import Image
    buf = BytesIO(img_data)
    img = Image.open(buf)
    if img.mode != "RGBA":
        img = img.convert("RGBA")
    rgb_data = img.tobytes("raw", "BGRA")
    w, h = img.size
    log("rgb_to_bitmap(%s) image size=%s, BGR buffer=%i bytes",
        Ellipsizer(rgb_data), img.size, len(rgb_data))
    header = BITMAPINFOHEADER()
    memset(byref(header), 0, sizeof(BITMAPINFOHEADER))
    header.biSize = sizeof(BITMAPINFOHEADER)
    header.biWidth = w
    header.biHeight = -h
    header.biPlanes = 1
    header.biBitCount = 32
    header.biCompression = BI_RGB
    header.biSizeImage = 0
    header.biXPelsPerMeter = 10
    header.biYPelsPerMeter = 10
    bitmapinfo = BITMAPINFO()
    bitmapinfo.bmiColors = 0
    memmove(byref(bitmapinfo.bmiHeader), byref(header), sizeof(BITMAPINFOHEADER))
    # noinspection PyTypeChecker
    buftype = c_char * len(rgb_data)
    # noinspection PyCallingNonCallable
    rgb_buf = buftype()
    rgb_buf.value = rgb_data
    pbuf = cast(byref(rgb_buf), c_void_p)
    hdc = GetDC(None)
    try:
        CBM_INIT = 4
        bitmap = CreateDIBitmap(hdc, byref(header), CBM_INIT, pbuf, byref(bitmapinfo), win32con.DIB_RGB_COLORS)
    finally:
        ReleaseDC(None, hdc)
    return bitmap


class Win32ClipboardProxy(ClipboardProxyCore):
    def __init__(self, window, selection, send_clipboard_request_handler, send_clipboard_token_handler):
        self.window = window
        self.send_clipboard_request_handler = send_clipboard_request_handler
        self.send_clipboard_token_handler = send_clipboard_token_handler
        super().__init__(selection)

    def with_clipboard_lock(self, success_callback, failure_callback, retries=RETRY, delay=DELAY):
        with_clipboard_lock(self.window, success_callback, failure_callback, retries=retries, delay=delay)

    def clear(self):
        self.with_clipboard_lock(self.empty_clipboard, clear_error)

    def empty_clipboard(self):
        r = EmptyClipboard()
        log("EmptyClipboard()=%s", r)
        return True

    def do_emit_token(self):
        if not self._greedy_client:
            # send just the token
            self.send_clipboard_token_handler(self)
            return

        # greedy clients want data with the token,
        # so we have to get the clipboard lock

        def send_token(formats: Sequence[int]) -> None:
            # default target:
            target = "UTF8_STRING"
            if formats:
                tnames = format_names(formats)
                if win32con.CF_UNICODETEXT in formats:
                    target = "UTF8_STRING"
                elif win32con.CF_TEXT in formats or win32con.CF_OEMTEXT in formats:
                    target = "STRING"
                elif "PNG" in tnames:
                    target = "image/png"

            def got_contents(dtype: str, dformat: int, data: Any) -> None:
                packet_data = ([target], (target, dtype, dformat, data))
                self.send_clipboard_token_handler(self, packet_data)

            self.get_contents(target, got_contents)

        def got_clipboard_lock():
            fmts = get_clipboard_formats()
            log("do_emit_token() formats=%s", format_names(fmts))
            send_token(fmts)
            return True

        def errback(errmsg=None):
            # nothing we can do!
            log("not sending the token since we failed to get the clipboard lock: %s", errmsg)

        self.with_clipboard_lock(got_clipboard_lock, errback)

    def get_contents(self, target: str, got_contents: ClipboardCallback):
        log("get_contents%s", (target, got_contents))
        if target == "TARGETS":
            def got_clipboard_lock() -> bool:
                formats = get_clipboard_formats()
                fnames = format_names(formats)
                targets = []
                if win32con.CF_UNICODETEXT in formats:
                    targets += ["text/plain;charset=utf-8", "UTF8_STRING", "CF_UNICODETEXT"]
                if win32con.CF_TEXT in formats or win32con.CF_OEMTEXT in formats:
                    targets += ["TEXT", "STRING", "text/plain"]
                # if any(x in fnames for x in ("CF_DIB", "CF_BITMAP", "CF_DIBV5")):
                if "CF_DIBV5" in fnames:
                    targets += ["image/png", "image/jpeg"]
                log("targets(%s)=%s", csv(fnames), csv(targets))
                got_contents("ATOM", 32, targets)
                return True

            def lockerror(_message) -> None:
                # assume text:
                got_contents("ATOM", 32, ["TEXT", "STRING", "text/plain", "text/plain;charset=utf-8", "UTF8_STRING"])

            self.with_clipboard_lock(got_clipboard_lock, lockerror)
            return

        def nodata(*args) -> None:
            log("nodata%s", args)
            got_contents(target, 8, b"")

        if target in ("image/png", "image/jpeg"):
            def got_image(img_data, trusted=False):
                log("got_image(%i bytes)", len(img_data))
                img_data = self.filter_data(dtype=target, dformat=8, data=img_data, trusted=trusted)
                got_contents(target, 8, img_data)

            img_format = target.split("/")[-1].upper()  # ie: "PNG" or "JPEG"
            self.get_clipboard_image(img_format, got_image, nodata)
            return
        if target not in ("TEXT", "STRING", "text/plain", "text/plain;charset=utf-8", "UTF8_STRING"):
            # we don't know how to handle this target,
            # return an empty response:
            nodata()
            return

        def got_text(text) -> None:
            log("got_text(%s)", Ellipsizer(text))
            got_contents(target, 8, text)

        def errback(error_text=""):
            log("errback(%s)", error_text)
            if error_text:
                log.warn("Warning: failed to get clipboard data as text")
                log.warn(" %s", error_text)
            got_contents(target, 8, b"")

        utf8 = target.lower().find("utf") >= 0
        self.get_clipboard_text(utf8, got_text, errback)

    def get_clipboard_image(self, img_format, got_image, errback):
        def got_clipboard_lock() -> bool:
            if COMPRESSED_IMAGES:
                fmt_name = LPCSTR(img_format.upper().encode("latin1") + b"\0")  # ie: "PNG"
                fmt = RegisterClipboardFormatA(fmt_name)
                if fmt:
                    data_handle = GetClipboardData(fmt)
                    if data_handle:
                        size = GlobalSize(data_handle)
                        data = GlobalLock(data_handle)
                        log("GetClipboardData(%s)=%#x size=%s, data=%#x",
                            img_format.upper(), data_handle, size, data)
                        if data and size:
                            try:
                                cdata = (c_char * size).from_address(data)
                            finally:
                                GlobalUnlock(data)
                            got_image(bytes(cdata), False)
                            return True

            data_handle = GetClipboardData(win32con.CF_DIBV5)
            log("CF_BITMAP=%s", data_handle)
            data = GlobalLock(data_handle)
            if not data:
                log("failed to lock data handle %#x (may try again)", data_handle)
                return False
            try:
                header = cast(data, PBITMAPV5HEADER).contents
                offset = header.bV5Size + header.bV5ClrUsed * 4
                w, h = header.bV5Width, abs(header.bV5Height)
                bits = header.bV5BitCount
                log("offset=%s, width=%i, height=%i, compression=%s",
                    offset, w, h, BI_FORMATS.get(header.bV5Compression, header.bV5Compression))
                log("planes=%i, bitcount=%i", header.bV5Planes, bits)
                log("colorspace=%s", COLOR_PROFILES.get(header.bV5CSType, header.bV5CSType))
                # if header.bV5Compression in (BI_JPEG, BI_PNG):
                #    pass
                if header.bV5Compression != BI_RGB:
                    errback("cannot handle %r compression yet" % BI_FORMATS.get(header.bV5Compression, "unknown"))
                    return True
                if bits == 24:
                    save_format = "RGB"
                    rgb_format = "BGR"
                    stride = roundup(w * 3, 4)
                elif bits == 32:
                    save_format = "RGBA"
                    rgb_format = "BGRA"
                    stride = w * 4
                else:
                    errback("cannot handle image data with %i bits per pixel yet" % bits)
                    return True
                img_size = stride * h
                # noinspection PyTypeChecker
                rgb_data = (c_char * img_size).from_address(data + offset)
                from PIL import Image, ImageOps
                img = Image.frombytes(save_format, (w, h), rgb_data, "raw", rgb_format, stride, 1)
                if header.bV5Height > 0:
                    img = ImageOps.flip(img)
                buf = BytesIO()
                img.save(buf, format=save_format)
                data = buf.getvalue()
                buf.close()
                got_image(data, True)
                return True
            finally:
                GlobalUnlock(data)

        self.with_clipboard_lock(got_clipboard_lock, errback)

    def got_token(self, targets, target_data=None, claim=True, _synchronous_client=False) -> None:
        # the remote end now owns the clipboard
        self.cancel_emit_token()
        if not self._enabled:
            return
        self._got_token_events += 1
        log("got token, selection=%s, targets=%s, target data=%s, claim=%s, can-receive=%s",
            self._selection, targets, Ellipsizer(target_data), claim, self._can_receive)
        if self._can_receive:
            self.targets = _filter_targets(targets or ())
            self.target_data = target_data or {}
            if targets:
                self.got_contents("TARGETS", "ATOM", 32, targets)
            if target_data:
                for target, td_def in target_data.items():
                    dtype, dformat, data = td_def
                    dtype = bytestostr(dtype)
                    self.got_contents(target, dtype, dformat, data)
            # since we claim to be greedy
            # the peer should have sent us the target and target_data,
            # if not then request it:
            if not targets:
                self.send_clipboard_request_handler(self, self._selection, "TARGETS")
        if not claim:
            log("token packet without claim, not setting the token flag")
            return
        self._have_token = True
        if self._can_receive:
            self.claim()

    def got_contents(self, target: str, dtype="", dformat=0, data=b""):
        # if this is the special target 'TARGETS', cache the result:
        if target == "TARGETS" and dtype == "ATOM" and dformat == 32:
            self.targets = _filter_targets(data)
            # TODO: tell system what targets we have
            log("got_contents: tell OS we have %s", csv(self.targets))
            image_formats = tuple(x for x in ("image/png", "image/jpeg") if x in self.targets)
            if image_formats:
                # request it:
                self.send_clipboard_request_handler(self, self._selection, image_formats[0])
        elif dformat == 8 and dtype in TEXT_TARGETS:
            log("we got a byte string: %s", Ellipsizer(data))
            if dtype.lower().find("utf8") >= 0:
                text = data.decode("utf8")
            else:
                text = bytestostr(data)
            self.set_clipboard_text(text)
        elif dformat == 8 and dtype.startswith("image/"):
            img_format = dtype.split("/")[-1]  # ie: 'png'
            self.set_clipboard_image(img_format, data)
        else:
            log("no handling: target=%s, dtype=%s, dformat=%s, data=%s",
                target, dtype, dformat, Ellipsizer(data))

    def set_clipboard_image(self, img_format, img_data):
        image_formats = {}
        if COMPRESSED_IMAGES:
            # first save it as binary compressed data:
            fmt_name = LPCSTR(img_format.upper().encode("latin1") + b"\0")  # ie: "PNG"
            fmt = RegisterClipboardFormatA(fmt_name)
            if fmt:
                l = len(img_data)
                # noinspection PyTypeChecker
                buftype = c_char * l
                # noinspection PyCallingNonCallable
                buf = buftype()
                buf.value = img_data
                pbuf = cast(byref(buf), c_void_p)
                data_handle = GlobalAlloc(GMEM_MOVEABLE, l)
                if not data_handle:
                    log.error("Error: failed to allocate %i bytes of global memory", l)
                    return True
                data = GlobalLock(data_handle)
                if not data:
                    log("failed to lock data handle %#x (may try again)", data_handle)
                    return False
                log("got data handle lock %#x for %i bytes of '%s' data", data, l, img_format)
                try:
                    memmove(data, pbuf, l)
                finally:
                    GlobalUnlock(data)
                image_formats[fmt] = data_handle
        bitmap = rgb_to_bitmap(img_data)
        image_formats[win32con.CF_BITMAP] = bitmap
        self.do_set_clipboard_image(image_formats)

    def do_set_clipboard_image(self, image_formats):
        def got_clipboard_lock():
            EmptyClipboard()
            c = 0
            for fmt, handle in image_formats.items():
                log("do_set_clipboard_image: %s", format_name(fmt))
                r = SetClipboardData(fmt, handle)
                if not r:
                    e = WinError(GetLastError())
                    log("SetClipboardData(%s, %#x)=%s (%s)", format_name(fmt), handle, r, e)
                else:
                    c += 1
            return bool(c)

        def nolock(*_args):
            log.warn("Warning: failed to copy image data to the clipboard")

        self.with_clipboard_lock(got_clipboard_lock, nolock)

    def get_clipboard_text(self, utf8, callback, errback):
        def get_text() -> bool:
            formats = get_clipboard_formats()
            matching = []
            for u in (utf8, not utf8):
                if u:
                    fmts = [win32con.CF_UNICODETEXT]
                else:
                    fmts = [win32con.CF_TEXT, win32con.CF_OEMTEXT]
                matching += [fmt for fmt in formats if fmt in fmts]
            log("supported formats: %s (prefer utf8: %s)", csv(format_names(matching)), utf8)
            if not matching:
                log("no supported formats, only: %s", csv(format_names(formats)))
                errback()
                return True
            data_handle = None
            fmt = None
            for fmt in matching:
                data_handle = GetClipboardData(fmt)
                log("GetClipboardData(%s)=%#x", format_name(fmt), data_handle or 0)
                if data_handle:
                    break
            if not data_handle:
                log("no valid data handle using %s (may try again)", csv(format_names(matching)))
                return False
            data = GlobalLock(data_handle)
            if not data:
                log("failed to lock data handle %#x (may try again)", data_handle)
                return False
            log("got data handle lock %#x for format '%s'", data, format_name(fmt))
            try:
                if fmt == win32con.CF_UNICODETEXT:
                    try:
                        v = w_to_utf8(data)
                    except Exception as e:
                        log("w_to_utf8(..)", exc_info=True)
                        errback(str(e))
                        return True
                    callback(v)
                    return True
                # CF_TEXT or CF_OEMTEXT:
                astr = cast(data, LPCSTR)
                s = astr.value.decode("latin1")
                if CONVERT_LINE_ENDINGS:
                    s = s.replace("\r\n", "\n")
                b = s.encode("latin1")
                ulen = len(b)
                if ulen > MAX_CLIPBOARD_PACKET_SIZE:
                    errback("text data is too large: %i characters" % ulen)
                    return True
                log("got %i bytes of TEXT data: %s", len(b), Ellipsizer(b))
                callback(b)
                return True
            finally:
                GlobalUnlock(data)

        self.with_clipboard_lock(get_text, errback)

    def set_clipboard_text(self, text: str):
        # convert to wide char
        # get the length in wide chars:
        if CONVERT_LINE_ENDINGS:
            text = text.replace("\n", "\r\n").encode("utf8")
        wlen = MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, text, len(text), None, 0)
        if not wlen:
            set_err("failed to prepare to convert to wide char")
            return True
        log("MultiByteToWideChar wlen=%i", wlen)
        # allocate some memory for it:
        l = (wlen + 1) * 2
        buf = GlobalAlloc(GMEM_MOVEABLE, l)
        if not buf:
            set_err("failed to allocate %i bytes of global memory" % l)
            return True
        log("GlobalAlloc buf=%#x", buf)
        locked = GlobalLock(buf)
        if not locked:
            set_err("failed to lock buffer %#x" % buf)
            GlobalFree(buf)
            return True
        try:
            locked_buf = cast(locked, LPWSTR)
            wlen = MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, text, len(text), locked_buf, wlen)
            if not wlen:
                set_err("failed to convert to wide char")
                GlobalFree(buf)
                return True
        finally:
            GlobalUnlock(locked)
        # we're going to alter the clipboard ourselves,
        # ignore messages until we're done,
        # this may take a while so ensure we do unblock eventually:
        long_timeout = GLib.timeout_add(2000, self.remove_block)
        self._block_owner_change = long_timeout

        def cleanup() -> None:
            if self._block_owner_change == long_timeout:
                # now safe to re-schedule and run now in the UI thread:
                self.cancel_unblock()
                self._block_owner_change = GLib.idle_add(self.remove_block)

        def set_clipboard_data() -> bool:
            r = EmptyClipboard()
            log("EmptyClipboard()=%s", r)
            if not r:
                set_err("failed to empty the clipboard")
                return False
            r = SetClipboardData(win32con.CF_UNICODETEXT, buf)
            if not r:
                e = WinError(GetLastError())
                log("SetClipboardData(CF_UNICODETEXT, %i chars)=%s (%s)", wlen, r, e)
                return False
            log("SetClipboardData(CF_UNICODETEXT, %i chars)=%s", wlen, r)
            cleanup()
            return True

        def set_clipboard_error(error_text="") -> None:
            log("set_clipboard_error(%s)", error_text)
            if error_text:
                log.warn("Warning: failed to set clipboard data")
                log.warn(" %s", error_text)
            cleanup()

        self.with_clipboard_lock(set_clipboard_data, set_clipboard_error)

    def __repr__(self):
        return "Win32ClipboardProxy"


class Win32Clipboard(ClipboardTimeoutHelper):
    """
        Use Native win32 API to access the clipboard
    """

    def __init__(self, send_packet_cb, progress_cb=noop, **kwargs):
        self.init_window()
        super().__init__(send_packet_cb, progress_cb, **kwargs)

    def init_window(self) -> None:
        log("Win32Clipboard.init_window() creating clipboard window class and instance")
        self.wndclass = WNDCLASSEX()
        self.wndclass.cbSize = sizeof(WNDCLASSEX)
        self.wndclass.lpfnWndProc = WNDPROC(self.wnd_proc)
        self.wndclass.style = win32con.CS_GLOBALCLASS
        self.wndclass.hInstance = GetModuleHandleA(0)
        self.wndclass.lpszClassName = CLIPBOARD_WINDOW_CLASS_NAME
        self.wndclass_handle = RegisterClassExW(byref(self.wndclass))
        log("RegisterClassExA(%s)=%#x", self.wndclass.lpszClassName, self.wndclass_handle)
        if self.wndclass_handle == 0:
            raise WinError()
        style = win32con.WS_CAPTION  # win32con.WS_OVERLAPPED
        self.window = CreateWindowExW(0, self.wndclass_handle, "Clipboard", style,
                                      0, 0, win32con.CW_USEDEFAULT, win32con.CW_USEDEFAULT,
                                      win32con.HWND_MESSAGE, 0, self.wndclass.hInstance, None)
        log("clipboard window=%#x", self.window)
        if not self.window:
            raise WinError()
        if not AddClipboardFormatListener(self.window):
            log.warn("Warning: failed to setup clipboard format listener")
            log.warn(" %s", get_last_error())

    def wnd_proc(self, hwnd: int, msg: int, wparam: int, lparam: int) -> int:
        r = DefWindowProcW(hwnd, msg, wparam, lparam)
        if msg not in CLIPBOARD_EVENTS:
            return r
        owner = GetClipboardOwner()
        log("clipboard event: %s, current owner: %s",
            CLIPBOARD_EVENTS.get(msg), get_owner_info(owner, self.window))
        if msg == WM_CLIPBOARDUPDATE and owner != self.window:
            owner = GetClipboardOwner()
            owner_info = get_owner_info(owner, self.window)
            if is_blocklisted(owner_info):
                # ie: don't try to sync from VirtualBox
                log("CLIPBOARDUPDATE coming from '%s' ignored", owner_info)
                return r
            min_delay = 500 * int(is_syncdelay(owner_info))
            log("CLIPBOARDUPDATE coming from '%s', min_delay=%i", owner_info, min_delay)
            for proxy in self._clipboard_proxies.values():
                if not proxy._block_owner_change:
                    proxy.schedule_emit_token(min_delay)
        return r

    def cleanup(self) -> None:
        super().cleanup()
        self.cleanup_window()

    def cleanup_window(self) -> None:
        w = self.window
        if w:
            self.window = None
            RemoveClipboardFormatListener(w)
            DestroyWindow(w)
        wch = self.wndclass_handle
        if wch:
            self.wndclass = None
            self.wndclass_handle = None
            UnregisterClassW(CLIPBOARD_WINDOW_CLASS_NAME, GetModuleHandleA(0))

    def make_proxy(self, selection: str) -> Win32ClipboardProxy:
        proxy = Win32ClipboardProxy(self.window, selection,
                                    self._send_clipboard_request_handler, self._send_clipboard_token_handler)
        proxy.set_want_targets(self._want_targets)
        proxy.set_direction(self.can_send, self.can_receive)
        return proxy

    ############################################################################
    # just pass ATOM targets through
    # (we use them internally as strings)
    ############################################################################
    def _munge_wire_selection_to_raw(self, encoding, dtype, dformat, data) -> bytes | str:
        if encoding == "atoms":
            return _filter_targets(data)
        return super()._munge_wire_selection_to_raw(encoding, dtype, dformat, data)

    def _munge_raw_selection_to_wire(self, target, dtype, dformat, data) -> tuple[Any, Any]:
        if dtype == "ATOM":
            assert isinstance(data, (tuple, list))
            return "atoms", _filter_targets(data)
        return super()._munge_raw_selection_to_wire(target, dtype, dformat, data)


def set_err(msg: str) -> None:
    log.warn("Warning: cannot set clipboard value")
    log.warn(" %s", msg)


def clear_error(error_text="") -> None:
    log.warn("Warning: failed to clear the clipboard")
    if error_text:
        log.warn(" %s", error_text)



# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


from xpra.x11.bindings.xlib cimport Display, Atom

cdef class X11CoreBindingsInstance:
    cdef Display * display
    cdef char * display_name
    cdef Atom xatom(self, str_or_int)
    cdef Atom str_to_atom(self, atomstr)
#    def get_error_text(self, code)



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from enum import IntEnum
from typing import TypeAlias


class ExitCode(IntEnum):
    OK = 0
    CONNECTION_LOST = 1
    TIMEOUT = 2
    PASSWORD_REQUIRED = 3
    PASSWORD_FILE_ERROR = 4
    INCOMPATIBLE_VERSION = 5
    ENCRYPTION = 6
    FAILURE = 7
    SSH_FAILURE = 8
    PACKET_FAILURE = 9
    MMAP_TOKEN_FAILURE = 10
    NO_AUTHENTICATION = 11
    UNSUPPORTED = 12
    REMOTE_ERROR = 13
    INTERNAL_ERROR = 14
    FILE_TOO_BIG = 15
    SSL_FAILURE = 16
    SSH_KEY_FAILURE = 17
    CONNECTION_FAILED = 18
    SSL_CERTIFICATE_VERIFY_FAILURE = 19
    NO_DISPLAY = 20
    SERVER_ALREADY_EXISTS = 21
    SOCKET_CREATION_ERROR = 22
    VFB_ERROR = 23
    FILE_NOT_FOUND = 24
    UPGRADE = 25
    IO_ERROR = 26
    NO_DATA = 27
    AUTHENTICATION_FAILED = 28
    DEVICE_NOT_FOUND = 29
    OPENGL_UNSAFE = 30
    COMPONENT_MISSING = 31


ExitValue: TypeAlias = ExitCode | int


def exit_str(code) -> str:
    try:
        return ExitCode(code).name
    except ValueError:
        return f"unknown error {code}"


RETRY_EXIT_CODES: list[ExitCode] = [
    ExitCode.CONNECTION_LOST,
    ExitCode.PACKET_FAILURE,
    ExitCode.UPGRADE,
]



# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Final
from collections.abc import Callable, Sequence

from xpra.os_util import gi_import
from xpra.util.env import first_time
from xpra.common import WORKSPACE_UNSET, WORKSPACE_ALL
from xpra.gtk.error import xsync, xlog
from xpra.x11.models.core import CoreX11WindowModel, Above, RESTACKING_STR
from xpra.x11.bindings.window import X11WindowBindings, constants
from xpra.server.window.content_guesser import guess_content_type, get_content_type_properties, GUESS_CONTENT
from xpra.server.background_worker import add_work_item
from xpra.x11.gtk.bindings import get_pywindow, get_pyatom
from xpra.log import Logger

log = Logger("x11", "window")
workspacelog = Logger("x11", "window", "workspace")
metalog = Logger("x11", "window", "metadata")

GObject = gi_import("GObject")

X11Window = X11WindowBindings()

# _NET_WM_STATE:
_NET_WM_STATE_REMOVE = 0
_NET_WM_STATE_ADD = 1
_NET_WM_STATE_TOGGLE = 2
STATE_STRING: dict[int, str] = {
    _NET_WM_STATE_REMOVE: "REMOVE",
    _NET_WM_STATE_ADD: "ADD",
    _NET_WM_STATE_TOGGLE: "TOGGLE",
}
WithdrawnState: Final[int] = constants["WithdrawnState"]
IconicState: Final[int] = constants["IconicState"]
NormalState: Final[int] = constants["NormalState"]
WM_STATE_NAME: dict[int, str] = {
    IconicState: "Iconic",
    NormalState: "Normal",
    WithdrawnState: "Withdrawn",
}

# add user-friendly workspace logging:
WORKSPACE_STR: dict[int, str] = {
    WORKSPACE_UNSET: "UNSET",
    WORKSPACE_ALL: "ALL",
}


def workspacestr(w) -> str:
    return WORKSPACE_STR.get(w, str(w))


class BaseWindowModel(CoreX11WindowModel):
    """
        Extends CoreX11WindowModel to add properties
        that we find on real X11 windows (as opposed to trays).
        Also wraps access to _NET_WM_STATE using simpler gobject booleans.
    """
    __common_properties__ = CoreX11WindowModel.__common_properties__.copy()
    __common_properties__ |= {
        # from WM_TRANSIENT_FOR
        "transient-for": (
            GObject.TYPE_PYOBJECT,
            "Transient for (or None)", "",
            GObject.ParamFlags.READABLE,
        ),
        # from _NET_WM_WINDOW_OPACITY
        "opacity": (
            GObject.TYPE_INT64,
            "Opacity", "",
            -1, 0xffffffff, -1,
            GObject.ParamFlags.READABLE,
        ),
        # from WM_HINTS.window_group
        "group-leader": (
            GObject.TYPE_PYOBJECT,
            "Window group leader as a pair: (xid, gdk window)", "",
            GObject.ParamFlags.READABLE,
        ),
        # from WM_HINTS.urgency or _NET_WM_STATE
        "attention-requested": (
            GObject.TYPE_BOOLEAN,
            "Urgency hint from client, or us", "",
            False,
            GObject.ParamFlags.READWRITE,
        ),
        # from WM_HINTS.input or WM_TAKE_FOCUS
        "can-focus": (
            GObject.TYPE_BOOLEAN,
            "Does this window ever accept keyboard input?", "",
            True,
            GObject.ParamFlags.READWRITE,
        ),
        # from _NET_WM_BYPASS_COMPOSITOR
        "bypass-compositor": (
            GObject.TYPE_INT,
            "hint that the window would benefit from running uncomposited ", "",
            0, 2, 0,
            GObject.ParamFlags.READABLE,
        ),
        # from _NET_WM_FULLSCREEN_MONITORS
        "fullscreen-monitors": (
            GObject.TYPE_PYOBJECT,
            "list of 4 monitor indices indicating the top, bottom, left, and right edges"
            " of the window when the fullscreen state is enabled", "",
            GObject.ParamFlags.READABLE,
        ),
        # from _NET_WM_STRUT_PARTIAL or _NET_WM_STRUT
        "strut": (
            GObject.TYPE_PYOBJECT,
            "Struts requested by window, or None", "",
            GObject.ParamFlags.READABLE,
        ),
        # for our own use:
        "content-type": (
            GObject.TYPE_PYOBJECT,
            "What type of content is shown in this window", "",
            GObject.ParamFlags.READABLE,
        ),
        # from _XPRA_QUALITY
        "quality": (
            GObject.TYPE_INT,
            "Quality", "",
            -1, 100, -1,
            GObject.ParamFlags.READABLE,
        ),
        # from _XPRA_SPEED
        "speed": (
            GObject.TYPE_INT,
            "Speed", "",
            -1, 100, -1,
            GObject.ParamFlags.READABLE,
        ),
        # from _XPRA_ENCODING
        "encoding": (
            GObject.TYPE_PYOBJECT,
            "Encoding", "",
            GObject.ParamFlags.READABLE,
        ),
        # from _NET_WM_DESKTOP
        "workspace": (
            GObject.TYPE_UINT,
            "The workspace this window is on", "",
            0, 2 ** 32 - 1, WORKSPACE_UNSET,
            GObject.ParamFlags.READWRITE,
        ),
        # set initially only by the window model class
        # (derived from XGetWindowAttributes.override_redirect)
        "override-redirect": (
            GObject.TYPE_BOOLEAN,
            "Is the window of type override-redirect", "",
            False,
            GObject.ParamFlags.READABLE,
        ),
        # from _NET_WM_WINDOW_TYPE
        "window-type": (
            GObject.TYPE_PYOBJECT,
            "Window type",
            "NB, most preferred comes first, then fallbacks",
            GObject.ParamFlags.READABLE,
        ),
        # this value is synced to "_NET_WM_STATE":
        "state": (
            GObject.TYPE_PYOBJECT,
            "State, as per _NET_WM_STATE", "",
            GObject.ParamFlags.READABLE,
        ),
        # all the attributes below are virtual attributes from WM_STATE:
        "modal": (
            GObject.TYPE_PYOBJECT,
            "Modal", "",
            GObject.ParamFlags.READWRITE,
        ),
        "fullscreen": (
            GObject.TYPE_BOOLEAN,
            "Fullscreen-ness of window", "",
            False,
            GObject.ParamFlags.READWRITE,
        ),
        "focused": (
            GObject.TYPE_BOOLEAN,
            "Is the window focused", "",
            False,
            GObject.ParamFlags.READWRITE,
        ),
        "maximized": (
            GObject.TYPE_BOOLEAN,
            "Is the window maximized", "",
            False,
            GObject.ParamFlags.READWRITE,
        ),
        "above": (
            GObject.TYPE_BOOLEAN,
            "Is the window on top of most windows", "",
            False,
            GObject.ParamFlags.READWRITE),
        "below": (
            GObject.TYPE_BOOLEAN,
            "Is the window below most windows", "",
            False,
            GObject.ParamFlags.READWRITE,
        ),
        "shaded": (
            GObject.TYPE_BOOLEAN,
            "Is the window shaded", "",
            False,
            GObject.ParamFlags.READWRITE,
        ),
        "skip-taskbar": (
            GObject.TYPE_BOOLEAN,
            "Should the window be included on a taskbar", "",
            False,
            GObject.ParamFlags.READWRITE,
        ),
        "skip-pager": (
            GObject.TYPE_BOOLEAN,
            "Should the window be included on a pager", "",
            False,
            GObject.ParamFlags.READWRITE,
        ),
        "sticky": (
            GObject.TYPE_BOOLEAN,
            "Is the window's position fixed on the screen", "",
            False,
            GObject.ParamFlags.READWRITE,
        ),
    }
    _property_names = CoreX11WindowModel._property_names + [
        "transient-for", "fullscreen-monitors", "bypass-compositor",
        "group-leader", "window-type", "workspace", "strut", "opacity",
        "content-type",
        # virtual attributes:
        "fullscreen", "focused", "maximized", "above", "below", "shaded",
        "skip-taskbar", "skip-pager", "sticky",
    ]
    _dynamic_property_names = CoreX11WindowModel._dynamic_property_names + [
        "attention-requested", "content-type",
        "window-type", "workspace", "opacity",
        "fullscreen", "focused", "maximized", "above", "below", "shaded",
        "skip-taskbar", "skip-pager", "sticky",
        "quality", "speed", "encoding",
    ]
    _internal_property_names = CoreX11WindowModel._internal_property_names + ["state"]
    _initial_x11_properties = CoreX11WindowModel._initial_x11_properties + [
        "WM_TRANSIENT_FOR",
        "_NET_WM_WINDOW_TYPE",
        "_NET_WM_DESKTOP",
        "_NET_WM_FULLSCREEN_MONITORS",
        "_NET_WM_BYPASS_COMPOSITOR",
        "_NET_WM_STRUT",
        # redundant as it uses the same handler as _NET_WM_STRUT:
        "_NET_WM_STRUT_PARTIAL",
        "_NET_WM_WINDOW_OPACITY",
        "WM_HINTS",
        "_XPRA_CONTENT_TYPE",
        "_XPRA_QUALITY",
        "_XPRA_SPEED",
        "_XPRA_ENCODING",
    ]
    _DEFAULT_NET_WM_ALLOWED_ACTIONS: list[str] = [
        f"_NET_WM_ACTION_{x}" for x in (
            "CLOSE", "MOVE", "RESIZE", "FULLSCREEN",
            "MINIMIZE", "SHADE", "STICK",
            "MAXIMIZE_HORZ", "MAXIMIZE_VERT",
            "CHANGE_DESKTOP", "ABOVE", "BELOW",
        )
    ]
    _MODELTYPE = "Base"

    def __init__(self, xid: int):
        super().__init__(xid)
        self.last_unmap_serial = 0
        self._input_field = True  # The WM_HINTS input field
        if GUESS_CONTENT:
            # watch for changes to properties that are used to derive the content-type:
            for x in get_content_type_properties():
                if x in self.get_dynamic_property_names():
                    self.connect(f"notify::{x}", self._content_type_related_property_change)

    def serial_after_last_unmap(self, serial) -> bool:
        # "The serial member is set from the serial number reported in the protocol
        # but expanded from the 16-bit least-significant bits to a full 32-bit value"
        if serial > self.last_unmap_serial:
            return True
        # the serial can wrap around:
        if self.last_unmap_serial - serial >= 2 ** 15:
            return True
        return False

    def _read_initial_X11_properties(self) -> None:
        metalog("%s.read_initial_X11_properties()", self._MODELTYPE)
        self._updateprop("state", frozenset(self._read_wm_state()))
        super()._read_initial_X11_properties()
        self._update_content_type()

    def _guess_window_type(self) -> str:
        # query the X11 property directly,
        # in case the python property isn't set yet
        if not self.is_OR():
            transient_for = self.prop_get("WM_TRANSIENT_FOR", "window")
            if transient_for:
                # EWMH says that even if it's transient-for, we MUST check to
                # see if it's override-redirect (and if so treat as NORMAL).
                # But we wouldn't be here if this was override-redirect.
                # (OverrideRedirectWindowModel overrides this method)
                return "_NET_WM_WINDOW_TYPE_DIALOG"
        return "_NET_WM_WINDOW_TYPE_NORMAL"

    ################################
    # Actions
    ################################

    def move_to_workspace(self, workspace: int) -> None:
        # we send a message to ourselves, we could also just update the property
        current = self.get_property("workspace")
        if current == workspace:
            workspacelog("move_to_workspace(%s) unchanged", workspacestr(workspace))
            return
        workspacelog("move_to_workspace(%s) current=%s", workspacestr(workspace), workspacestr(current))
        with xlog:
            if workspace == WORKSPACE_UNSET:
                workspacelog("removing _NET_WM_DESKTOP property from window %#x", self.xid)
                self.prop_del("_NET_WM_DESKTOP")
            else:
                workspacelog("setting _NET_WM_DESKTOP=%s on window %#x", workspacestr(workspace), self.xid)
                self.prop_set("_NET_WM_DESKTOP", "u32", workspace)

    #########################################
    # Python objects synced to X11 properties
    #########################################

    def _sync_state(self, *_args) -> None:
        state = self.get_property("state")
        if "_NET_WM_STATE_HIDDEN" in state and "_NET_WM_STATE_FOCUSED" in state:
            # a window should not be focused and hidden at the same time:
            state = set(state)
            state.remove("_NET_WM_STATE_HIDDEN")
            state = frozenset(state)
        metalog("sync_state: setting _NET_WM_STATE=%s on %#x", state, self.xid)
        with xlog:
            self.prop_set("_NET_WM_STATE", ["atom"], state)

    def set_wm_state(self, state: int):
        metalog("set_wm_state(%s)", state)
        with xlog:
            self.prop_set("WM_STATE", "state", state)

    def _sync_iconic(self, *_args) -> None:
        if self.get("iconic"):
            self.set_wm_state(IconicState)
            self._state_update(add=("_NET_WM_STATE_HIDDEN", ), remove=("_NET_WM_STATE_FOCUSED", ))
        else:
            self.set_wm_state(NormalState)
            self._state_update(remove=("_NET_WM_STATE_HIDDEN", ))

    _py_property_handlers: dict[str, Callable] = dict(CoreX11WindowModel._py_property_handlers)
    _py_property_handlers |= {
        "state": _sync_state,
        "iconic": _sync_iconic,
    }

    #########################################
    # X11 properties synced to Python objects
    #########################################

    def _handle_transient_for_change(self) -> None:
        transient_for = self.prop_get("WM_TRANSIENT_FOR", "window")
        metalog("WM_TRANSIENT_FOR=%s", transient_for)
        # `transient-for` may be None
        window = get_pywindow(transient_for) if transient_for else None
        self._updateprop("transient-for", window)

    def _handle_window_type_change(self) -> None:
        window_types = self.prop_get("_NET_WM_WINDOW_TYPE", ["atom"])
        metalog("_NET_WM_WINDOW_TYPE=%s", window_types)
        if not window_types:
            window_type = self._guess_window_type()
            metalog("guessed window type=%s", window_type)
            # atom = Gdk.Atom.intern(window_type, False)
            window_types = [window_type]
        # normalize them (hide _NET_WM_WINDOW_TYPE prefix):
        window_types = [str(wt).replace("_NET_WM_WINDOW_TYPE_", "").replace("_NET_WM_TYPE_", "") for wt in window_types]
        self._updateprop("window-type", window_types)

    def _handle_workspace_change(self) -> None:
        workspace = self.prop_get("_NET_WM_DESKTOP", "u32", True)
        if workspace is None:
            workspace = WORKSPACE_UNSET
        workspacelog("_NET_WM_DESKTOP=%s for window %#x", workspacestr(workspace), self.xid)
        self._updateprop("workspace", workspace)

    def _handle_fullscreen_monitors_change(self) -> None:
        fsm = self.prop_get("_NET_WM_FULLSCREEN_MONITORS", ["u32"], True)
        metalog("_NET_WM_FULLSCREEN_MONITORS=%s", fsm)
        self._updateprop("fullscreen-monitors", fsm)

    def _handle_bypass_compositor_change(self) -> None:
        bypass = self.prop_get("_NET_WM_BYPASS_COMPOSITOR", "u32", True) or 0
        metalog("_NET_WM_BYPASS_COMPOSITOR=%s", bypass)
        self._updateprop("bypass-compositor", bypass)

    def _handle_wm_strut_change(self) -> None:
        strut = self.prop_get("_NET_WM_STRUT_PARTIAL", "strut-partial")
        metalog("_NET_WM_STRUT_PARTIAL=%s", strut)
        if strut is None:
            strut = self.prop_get("_NET_WM_STRUT", "strut")
            metalog("_NET_WM_STRUT=%s", strut)
        # Might be None:
        self._updateprop("strut", strut)

    def _handle_opacity_change(self) -> None:
        opacity = self.prop_get("_NET_WM_WINDOW_OPACITY", "u32", True) or -1
        metalog("_NET_WM_WINDOW_OPACITY=%s", opacity)
        self._updateprop("opacity", opacity)

    def _handle_wm_hints_change(self) -> None:
        with xsync:
            wm_hints = X11Window.getWMHints(self.xid)
        metalog("getWMHints(%#x)=%s", self.xid, wm_hints)
        if wm_hints is None:
            return
        # GdkWindow or None
        group_leader = None
        if "window_group" in wm_hints:
            xid = wm_hints.get("window_group")
            if xid:
                group_leader = get_pywindow(xid)
                if not group_leader:
                    log.error(f"Error group leader window {xid:x} not found")
        self._updateprop("group-leader", group_leader)
        self._updateprop("attention-requested", wm_hints.get("urgency", False))
        _input = wm_hints.get("input")
        metalog("wm_hints.input = %s", _input)
        # we only set this value once:
        # (input_field always starts as True, and we then set it to an int)
        if self._input_field is True and _input is not None:
            # keep the value as an int to differentiate from the start value:
            self._input_field = int(_input)
            self._update_can_focus()

    def _update_can_focus(self, *_args) -> None:
        can_focus = bool(self._input_field) or "WM_TAKE_FOCUS" in self.get_property("protocols")
        self._updateprop("can-focus", can_focus)

    def _content_type_related_property_change(self, *_args) -> None:
        self._update_content_type()

    def _handle_xpra_content_type_change(self) -> None:
        self._update_content_type()

    def _update_content_type(self) -> None:
        # watch for changes to properties that are used to derive the content-type:
        content_type = self.prop_get("_XPRA_CONTENT_TYPE", "latin1", True) or ""
        # the _XPRA_CONTENT_TYPE property takes precedence
        if not content_type and GUESS_CONTENT:
            # guess_content_type can take a while as it will load the xdg menu data
            # so call it as a work item.
            # multiple callers can be added to the sequential work queue,
            # the menu loader also has a lock.
            add_work_item(self.guess_content_type)
        metalog("_update_content_type() %s", content_type)
        self._updateprop("content-type", content_type)

    def guess_content_type(self) -> None:
        content_type = guess_content_type(self)
        if not content_type and self.is_tray():
            content_type = "picture"
        metalog(f"guess_content_type() {content_type=}")
        self._updateprop("content-type", content_type)

    def _handle_xpra_quality_change(self) -> None:
        quality = self.prop_get("_XPRA_QUALITY", "u32", True) or -1
        metalog("quality=%s", quality)
        self._updateprop("quality", max(-1, min(100, quality)))

    def _handle_xpra_speed_change(self) -> None:
        speed = self.prop_get("_XPRA_SPEED", "u32", True) or -1
        metalog("speed=%s", speed)
        self._updateprop("speed", max(-1, min(100, speed)))

    def _handle_xpra_encoding_change(self) -> None:
        encoding = self.prop_get("_XPRA_ENCODING", "latin1", True) or ""
        metalog("encoding=%s", encoding)
        self._updateprop("encoding", encoding)

    _x11_property_handlers: dict[str, Callable] = CoreX11WindowModel._x11_property_handlers.copy()
    _x11_property_handlers |= {
        "WM_TRANSIENT_FOR": _handle_transient_for_change,
        "_NET_WM_WINDOW_TYPE": _handle_window_type_change,
        "_NET_WM_DESKTOP": _handle_workspace_change,
        "_NET_WM_FULLSCREEN_MONITORS": _handle_fullscreen_monitors_change,
        "_NET_WM_BYPASS_COMPOSITOR": _handle_bypass_compositor_change,
        "_NET_WM_STRUT": _handle_wm_strut_change,
        "_NET_WM_STRUT_PARTIAL": _handle_wm_strut_change,
        "_NET_WM_WINDOW_OPACITY": _handle_opacity_change,
        "WM_HINTS": _handle_wm_hints_change,
        "_XPRA_CONTENT_TYPE": _handle_xpra_content_type_change,
        "_XPRA_QUALITY": _handle_xpra_quality_change,
        "_XPRA_SPEED": _handle_xpra_speed_change,
        "_XPRA_ENCODING": _handle_xpra_encoding_change,
    }

    #########################################
    # _NET_WM_STATE
    #########################################
    # A few words about _NET_WM_STATE are in order.  Basically, it is a set of
    # flags.  Clients are allowed to set the initial value of this X property
    # to anything they like, when their window is first mapped; after that,
    # though, only the window manager is allowed to touch this property.  So
    # we store its value (or at least, our idea as to its value, the X server
    # in principle could disagree) as the "state" property.  There are
    # basically two things we need to accomplish:
    #   1) Whenever our property is modified, we mirror that modification into
    #      the X server.  This is done by connecting to our own notify::state
    #      signal.
    #   2) As a more user-friendly interface to these state flags, we provide
    #      several boolean properties like "attention-requested".
    #      These are virtual boolean variables; they are actually backed
    #      directly by the "state" property, and reading/writing them in fact
    #      accesses the "state" set directly.  This is done by overriding
    #      do_set_property and do_get_property.
    _state_properties: dict[str, Sequence[str]] = {
        "attention-requested": ("_NET_WM_STATE_DEMANDS_ATTENTION",),
        "fullscreen": ("_NET_WM_STATE_FULLSCREEN",),
        "maximized": ("_NET_WM_STATE_MAXIMIZED_VERT", "_NET_WM_STATE_MAXIMIZED_HORZ"),
        "shaded": ("_NET_WM_STATE_SHADED",),
        "above": ("_NET_WM_STATE_ABOVE",),
        "below": ("_NET_WM_STATE_BELOW",),
        "sticky": ("_NET_WM_STATE_STICKY",),
        "skip-taskbar": ("_NET_WM_STATE_SKIP_TASKBAR",),
        "skip-pager": ("_NET_WM_STATE_SKIP_PAGER",),
        "modal": ("_NET_WM_STATE_MODAL",),
        "focused": ("_NET_WM_STATE_FOCUSED",),
    }
    _state_properties_reversed = {}
    for k, states in tuple(_state_properties.items()):
        for x in states:
            _state_properties_reversed[x] = k

    def _state_update(self, add=(), remove=()) -> None:
        curr = set(self.get_property("state"))
        was = set(curr)
        added = set(add) - curr
        curr |= added
        removed = set(remove) & curr
        curr -= removed
        metalog("state_update(%s, %s) was=%s, added=%s, removed=%s, new state=%s",
                add, remove, was, added, removed, curr)
        if added or removed:
            # note: _sync_state will update _NET_WM_STATE here:
            self._internal_set_property("state", frozenset(curr))
            self._state_notify(added | removed)

    def _state_notify(self, state_names: set[str]) -> None:
        notify_props: set[str] = set()
        for x in state_names:
            if x in self._state_properties_reversed:
                notify_props.add(self._state_properties_reversed[x])
        metalog("state notify(%s)=%s", state_names, notify_props)
        for x in tuple(notify_props):
            self.notify(x)

    def _state_isset(self, state_name: str) -> bool:
        return state_name in self.get_property("state")

    def _read_wm_state(self) -> list[str]:
        wm_state = self.prop_get("_NET_WM_STATE", ["atom"])
        metalog("read _NET_WM_STATE=%s", wm_state)
        return wm_state or []

    def do_set_property(self, pspec, value) -> None:
        # intercept state properties to route via update_state()
        if pspec.name in self._state_properties:
            # virtual property for WM_STATE:
            self.update_wm_state(pspec.name, bool(value))
            return
        super().do_set_property(pspec, value)

    def do_get_property(self, pspec) -> object:
        # intercept state properties to route via get_wm_state()
        if pspec.name in self._state_properties:
            # virtual property for WM_STATE:
            return self.get_wm_state(pspec.name)
        return super().do_get_property(pspec)

    def update_wm_state(self, prop: str, b: bool) -> None:
        state_names = self._state_properties.get(prop)
        if not state_names:
            raise ValueError(f"invalid window state {prop}")
        if b:
            self._state_update(add=state_names)
        else:
            self._state_update(remove=state_names)

    def get_wm_state(self, prop: str) -> bool:
        state_names = self._state_properties.get(prop)
        if not state_names:
            raise ValueError(f"invalid window state {prop}")
        # this is a virtual property for WM_STATE:
        # return True if any is set (only relevant for maximized)
        return any(self._state_isset(x) for x in state_names)

    #########################################
    # X11 Events
    #########################################

    def process_client_message_event(self, event) -> bool:
        # FIXME
        # Need to listen for:
        #   _NET_CURRENT_DESKTOP
        #   _NET_WM_PING responses
        # and maybe:
        #   _NET_WM_STATE (more fully)

        if event.message_type == "_NET_WM_STATE":
            def update_wm_state(prop: str) -> None:
                current = self.get_property(prop)
                mode = event.data[0]
                if mode == _NET_WM_STATE_ADD:
                    v = True
                elif mode == _NET_WM_STATE_REMOVE:
                    v = False
                elif mode == _NET_WM_STATE_TOGGLE:
                    v = not bool(current)
                else:
                    log.warn(f"Warning: invalid mode for _NET_WM_STATE: {mode}")
                    return
                log("process_client_message_event(%s) window %s=%s after %s (current state=%s)",
                    event, prop, v, STATE_STRING.get(mode, mode), current)
                if v != current:
                    self.update_wm_state(prop, v)

            atom1 = get_pyatom(event.data[1])
            log("_NET_WM_STATE: %s", atom1)
            if atom1 == "_NET_WM_STATE_FULLSCREEN":
                update_wm_state("fullscreen")
            elif atom1 == "_NET_WM_STATE_ABOVE":
                update_wm_state("above")
            elif atom1 == "_NET_WM_STATE_BELOW":
                update_wm_state("below")
            elif atom1 == "_NET_WM_STATE_SHADED":
                update_wm_state("shaded")
            elif atom1 == "_NET_WM_STATE_STICKY":
                update_wm_state("sticky")
            elif atom1 == "_NET_WM_STATE_SKIP_TASKBAR":
                update_wm_state("skip-taskbar")
            elif atom1 == "_NET_WM_STATE_SKIP_PAGER":
                update_wm_state("skip-pager")
                get_pyatom(event.data[2])
            elif atom1 in ("_NET_WM_STATE_MAXIMIZED_VERT", "_NET_WM_STATE_MAXIMIZED_HORZ"):
                atom2 = get_pyatom(event.data[2])
                # we only have one state for both, so we require both to be set:
                if atom1 != atom2 and atom2 in ("_NET_WM_STATE_MAXIMIZED_VERT", "_NET_WM_STATE_MAXIMIZED_HORZ"):
                    update_wm_state("maximized")
            elif atom1 == "_NET_WM_STATE_HIDDEN":
                log("ignoring 'HIDDEN' _NET_WM_STATE: %s", event)
                # we don't honour those because they make little sense, see:
                # https://mail.gnome.org/archives/wm-spec-list/2005-May/msg00004.html
            elif atom1 == "_NET_WM_STATE_MODAL":
                update_wm_state("modal")
            elif atom1 == "_NET_WM_STATE_DEMANDS_ATTENTION":
                update_wm_state("attention-requested")
            elif atom1 == "_KDE_NET_WM_STATE_SKIP_SWITCHER":
                if first_time(f"KDE-WM_STATE-{self.xid}"):
                    log.info(f"invalid KDE specific request {atom1} ignored, will not be logged again")
            else:
                log.info(f"Unhandled _NET_WM_STATE request: {atom1!r}")
                log.info(f" event {event!r}")
            return True
        if event.message_type == "WM_CHANGE_STATE":
            iconic = event.data[0]
            log("WM_CHANGE_STATE: %s, serial=%s, last unmap serial=%#x",
                WM_STATE_NAME.get(iconic, iconic), event.serial, self.last_unmap_serial)
            valid = iconic in (IconicState, NormalState)
            if valid and self.serial_after_last_unmap(event.serial) and not self.is_OR() and not self.is_tray():
                self._updateprop("iconic", iconic == IconicState)
            return True
        if event.message_type == "_NET_WM_MOVERESIZE":
            log("_NET_WM_MOVERESIZE: %s", event)
            self.emit("initiate-moveresize", event)
            return True
        if event.message_type == "_NET_ACTIVE_WINDOW":
            # to filter based on the source indication:
            # ACTIVE_WINDOW_SOURCE = tuple(int(x)
            #                              for x in os.environ.get("XPRA_ACTIVE_WINDOW_SOURCE", "0,1").split(","))
            # if event.data[0] in ACTIVE_WINDOW_SOURCE:
            log("_NET_ACTIVE_WINDOW: %s", event)
            self.set_active()
            self.emit("restack", Above, 0)
            return True
        if event.message_type == "_NET_WM_DESKTOP":
            workspace = int(event.data[0])
            # query the workspace count on the root window
            # since we cannot access Wm from here…
            ndesktops = int(self.root_prop_get("_NET_NUMBER_OF_DESKTOPS", "u32") or 0)
            workspacelog("received _NET_WM_DESKTOP: workspace=%s, number of desktops=%s",
                         workspacestr(workspace), ndesktops)
            if ndesktops > 0 and (workspace in (WORKSPACE_UNSET, WORKSPACE_ALL) or 0 <= workspace < ndesktops):
                self.move_to_workspace(workspace)
            else:
                workspacelog.warn("invalid _NET_WM_DESKTOP request: workspace=%s, number of desktops=%s",
                                  workspacestr(workspace), ndesktops)
            return True
        if event.message_type == "_NET_WM_FULLSCREEN_MONITORS":
            log("_NET_WM_FULLSCREEN_MONITORS: %s", event)
            # TODO: we should validate the indexes instead of copying them blindly!
            # TODO: keep track of source indication so we can forward that to the client
            N = 16  # FIXME: this is an arbitrary limit
            monitors = list(event.data[:4])
            if not all(0 <= x < N for x in monitors):
                log.warn("Warning: invalid list of _NET_WM_FULLSCREEN_MONITORS:%s - ignored", event.data)
                return False
            log("_NET_WM_FULLSCREEN_MONITORS: monitors=%s", monitors)
            self.prop_set("_NET_WM_FULLSCREEN_MONITORS", ["u32"], monitors)
            return True
        if event.message_type == "_NET_RESTACK_WINDOW":
            source = {1: "application", 2: "pager"}.get(event.data[0], "default (%s)" % event.data[0])
            sibling_window = event.data[1]
            log("%s sent to window %#x for sibling %#x from %s with detail=%s",
                event.message_type, event.window, sibling_window, source,
                RESTACKING_STR.get(event.detail, event.detail))
            self.emit("restack", event.detail, sibling_window)
            return True
        # TODO: maybe we should process _NET_MOVERESIZE_WINDOW here?
        # it may make sense to apply it to the client_window
        # whereas the code in WindowModel assumes there is a corral window
        # not handled:
        return super().process_client_message_event(event)



# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# authentication from a file containing a list of entries of the form:
# username|password|uid|gid|displays|env_options|session_options

from xpra.auth.common import SessionData, parse_uid, parse_gid
from xpra.auth.file_auth_base import log, FileAuthenticatorBase
from xpra.util.str_fn import hexstr
from xpra.util.parsing import parse_str_dict
from xpra.util.objects import typedict
from xpra.net.digest import verify_digest

AuthLine = tuple[str, str, int, int, list[str], dict[str, str], dict[str, str]]


def parse_auth_line(line: str) -> AuthLine:
    ldata = line.split("|")
    if len(ldata) < 2:
        raise ValueError(f"not enough fields: {len(ldata)}, minimum is 2")
    log(f"found {len(ldata)} fields")
    # parse fields:
    username = ldata[0]
    password = ldata[1]
    if len(ldata) >= 5:
        uid = parse_uid(ldata[2])
        gid = parse_gid(ldata[3])
        displays = ldata[4].split(",")
    else:
        # this will use the default value, usually "nobody":
        uid = parse_uid(None)
        gid = parse_gid(None)
        displays = []
    env_options: dict[str, str] = {}
    session_options: dict[str, str] = {}
    if len(ldata) >= 6:
        env_options = parse_str_dict(ldata[5], ";")
    if len(ldata) >= 7:
        session_options = parse_str_dict(ldata[6], ";")
    return username, password, uid, gid, displays, env_options, session_options


class Authenticator(FileAuthenticatorBase):
    CLIENT_USERNAME = True

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.sessions: SessionData | None = None

    def parse_filedata(self, data: str) -> dict[str, AuthLine]:
        if not data:
            return {}
        auth_data: dict[str, AuthLine] = {}
        i = 0
        for line in data.splitlines():
            i += 1
            line = line.strip()
            log(f"line {i}: {line!r}")
            if not line or line.startswith("#"):
                continue
            try:
                v = parse_auth_line(line)
                if v:
                    username = v[0]
                    if username in auth_data:
                        log.error(f"Error: duplicate entry for username {username!r} in {self.password_filename!r}")
                    else:
                        auth_data[username] = v
            except Exception as e:
                log("parsing error", exc_info=True)
                log.error(f"Error parsing password file {self.password_filename!r} at line {i}:")
                log.error(f" {line!r}")
                log.estr(e)
                continue
        log(f"parsed auth data from file {self.password_filename!r}: {auth_data}")
        return auth_data

    def get_auth_info(self) -> AuthLine | None:
        self.load_password_file()
        if not self.password_filedata:
            return None
        return self.password_filedata.get(self.username)

    def get_password(self) -> str:
        entry = self.get_auth_info()
        log(f"get_password() found entry={entry}")
        if entry is None:
            return ""
        return entry[1]

    def authenticate_hmac(self, caps: typedict) -> bool:
        challenge_response = caps.bytesget("challenge_response")
        client_salt = caps.bytesget("challenge_client_salt")
        log(f"multifile_auth.authenticate_hmac challenge-response={challenge_response!r}, client-salt={client_salt!r}")
        self.sessions = None
        if not self.salt:
            log.error("Error: illegal challenge response received - salt cleared or unset")
            return False
        # ensure this salt does not get re-used:
        salt = self.get_response_salt(client_salt)
        entry = self.get_auth_info()
        if entry is None:
            log.warn("Warning: authentication failed")
            log.warn(f" no password for {self.username!r} in {self.password_filename!r}")
            return False
        log("authenticate: auth-info(%s)=%s", self.username, entry)
        fpassword, uid, gid, displays, env_options, session_options = entry[1:]
        log("multifile authenticate_hmac password='%r', hex(salt)=%s", fpassword, hexstr(salt))
        if not verify_digest(self.digest, fpassword, salt, challenge_response):
            log.warn("Warning: %s challenge for '%s' does not match", self.digest, self.username)
            return False
        self.sessions = uid, gid, displays, env_options, session_options
        return True

    def get_sessions(self) -> SessionData | None:
        return self.sessions

    def __repr__(self):
        return "multi password file"



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>

import sys
import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
gi.require_version('Gdk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk   # pylint: disable=wrong-import-position @UnresolvedImport


def main():
	window = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
	window.connect("destroy", Gtk.main_quit)
	window.set_default_size(320, 200)
	window.set_border_width(20)

	entry = Gtk.Entry()
	entry.set_max_length(100)
	entry.set_width_chars(32)

	window.add(entry)
	window.show_all()
	Gtk.main()


if __name__ == "__main__":
	main()
	sys.exit(0)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys

from xpra.platform import platform_import


def load_menu() -> dict:
    return {}


def load_desktop_sessions() -> dict:
    return {}


def clear_cache() -> None:
    """ the Posix override has a cache to clear """


platform_import(globals(), "menu_helper", False,
                "load_menu",
                "load_desktop_sessions",
                "clear_cache",
                )


def main():
    import os
    from xpra.util.str_fn import print_nested_dict
    from xpra.log import consume_verbose_argv
    from xpra.platform import program_context  # pylint: disable=import-outside-toplevel
    with program_context("Menu-Helper", "Menu Helper"):
        consume_verbose_argv(sys.argv, "menu")

        def icon_fmt(icondata):
            return "%i bytes" % len(icondata)

        if len(sys.argv) > 1:
            for x in sys.argv[1:]:
                if os.path.isabs(x):
                    from xpra.codecs.icon_util import load_icon_from_file
                    v = load_icon_from_file(x)
                    print(f"load_icon_from_file({x})={v}")
        else:
            menu = load_menu()
            if menu:
                print()
                print("application menu:")
                print_nested_dict(menu, vformat={"IconData": icon_fmt})
            else:
                print("no application menu data found")
            # try desktop sessions:
            sessions = load_desktop_sessions()
            if sessions:
                print()
                print("session menu:")
                print_nested_dict(sessions, vformat={"IconData": icon_fmt})
            else:
                print("no session menu data found")
    return 0


if __name__ == "__main__":
    r = main()
    sys.exit(r)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import socket
from zeroconf import ServiceInfo, Zeroconf, __version__ as zeroconf_version

from xpra.log import Logger
from xpra.util.env import envbool, envint, first_time
from xpra.net.net_util import get_interfaces_addresses
from xpra.net.mdns import XPRA_TCP_MDNS_TYPE, XPRA_UDP_MDNS_TYPE
from xpra.net.net_util import get_iface

log = Logger("network", "mdns")
log("python-zeroconf version %s", zeroconf_version)

IPV6 = envbool("XPRA_ZEROCONF_IPV6", True)
IPV6_LO = envbool("XPRA_ZEROCONF_IPV6_LOOPBACK", False)
IPV4_LO = envint("XPRA_ZEROCONF_IPV4_LOOPBACK", 1)


LOOPBACK_AFAM = {
    "0.0.0.0": socket.AF_INET,
    "::": socket.AF_INET6,
}


def get_interface_index(host: str) -> str:
    # we don't use interface numbers with zeroconf,
    # so just return the interface name,
    # which is also unique
    return get_iface(host)


def inet_ton(af, addr):
    if af == socket.AF_INET:
        return socket.inet_aton(addr)
    return socket.inet_pton(af, addr)  # @UndefinedVariable


def txt_rec(text_dict) -> dict:
    # prevent zeroconf from mangling our ints into booleans:
    new_dict = {}
    for k, v in text_dict.items():
        if isinstance(v, int):
            new_dict[k] = str(v)
        else:
            new_dict[k] = v
    return new_dict


class ZeroconfPublishers:
    """
    Expose services via python zeroconf
    """

    def __init__(self, listen_on, service_name: str, service_type: str = XPRA_TCP_MDNS_TYPE, text_dict=None):
        log("ZeroconfPublishers%s", (listen_on, service_name, service_type, text_dict))
        self.services: list[ZeroconfPublisher] = []
        self.ports: dict[str, set[int]] = {}

        def add_address(host: str, port: int, af=socket.AF_INET):
            try:
                if af == socket.AF_INET6 and host.find("%"):
                    host = host.split("%")[0]
                try:
                    host = socket.gethostbyname(host)
                except Exception:
                    pass
                address = inet_ton(af, host)
            except Exception as e:
                log("inet_aton(%s)", host, exc_info=True)
                log.warn(f"Warning: cannot publish {service_name} records on {host!r}:")
                log.warn(" %s", e)
                return
            sn = service_name
            ports = set(self.ports.get(service_name, ()))
            log("add_address(%s, %s, %s) ports=%s", host, port, af, ports)
            ports.add(port)
            if len(ports) > 1:
                sn += f"-{len(ports)}"
            try:
                zp = ZeroconfPublisher(address, host, port, sn, service_type, text_dict)
            except Exception as e:
                log.warn(f"Warning: zeroconf API error for {service_name} on {host!r}:{port}:")
                log.warn(" %s", e)
            else:
                self.services.append(zp)
                self.ports[service_name] = ports

        for host, port in listen_on:
            if host.startswith("[") and host.endswith("]"):
                host = host[1:-1]
            if host in ("0.0.0.0", "::"):
                af = LOOPBACK_AFAM.get(host)
                if af == socket.AF_INET6 and not IPV6_LO:
                    if first_time(f"zeroconf-{host}"):
                        log.info(f"python-zeroconf: {host!r} IPv6 loopback address is not supported")
                        mode = (text_dict or {}).get("mode")
                        if mode:
                            log.info(f" unable to publish mDNS record for {mode} connections")
                        log("try XPRA_ZEROCONF_IPV6_LOOPBACK=1 to enable it at your own risk")
                    # means that IPV6 is False and "::" is not supported
                    # at time of writing, https://pypi.org/project/zeroconf/ says:
                    # _listening on localhost (::1) does not work. Help with understanding why is appreciated._
                    continue

                # annoying: we have to enumerate all interfaces
                iaddr = get_interfaces_addresses()

                def add_iface(iface: str, addresses: dict) -> None:
                    log("%r %s: %s", iface, af, addresses.get(af, {}))
                    for defs in addresses.get(af, ()):
                        addr = defs.get("addr")
                        if addr:
                            add_address(addr, port, af)

                log(f"interface addresses: {iaddr!r}")
                # ensure loopback is done last, as it may conflict:
                loopback = iaddr.pop("lo", {})
                for iface, addresses in iaddr.items():
                    add_iface(iface, addresses)
                # with IPV4_LO=2, always publish loopback,
                # with IPV4_LO=1, only publish it if we don't have other addresses:
                if loopback and (IPV4_LO == 2) or (IPV4_LO == 1 and not iaddr):
                    add_iface("lo", loopback)
                continue

            if host == "":
                host = "127.0.0.1"
            af = socket.AF_INET
            if host.find(":") >= 0:
                if IPV6:
                    af = socket.AF_INET6
                else:
                    host = "127.0.0.1"
            add_address(host, port, af)

    def start(self) -> None:
        for s in self.services:
            s.start()

    def stop(self) -> None:
        for s in self.services:
            s.stop()

    def update_txt(self, txt) -> None:
        for s in self.services:
            try:
                s.update_txt(txt)
            except TimeoutError:
                log(f"update_txt({txt})", exc_info=True)


class ZeroconfPublisher:
    def __init__(self, address, host: str, port: int,
                 service_name: str, service_type=XPRA_TCP_MDNS_TYPE,
                 text_dict=None):
        log("ZeroconfPublisher%s", (address, host, port, service_name, service_type, text_dict))
        self.address = address
        self.host = host
        self.port = port
        self.zeroconf = None
        self.service = None
        self.kwargs = {}
        self.registered = False
        try:
            # ie: service_name = localhost.localdomain :2 (ssl)
            parts = service_name.split(" ", 1)
            regname = parts[0].split(".")[0]
            if len(parts) == 2:
                regname += parts[1]
            regname = regname.replace(":", "-")
            regname = regname.replace(" ", "-")
            regname = regname.replace("(", "")
            regname = regname.replace(")", "")
            # ie: regname = localhost-2-ssl
            st = service_type + "local."
            regname += "." + st
            td = txt_rec(text_dict or {})
            self.kwargs = {
                "type_": st,  # "_xpra._tcp.local."
                "name": regname,
                "server": regname,
                "port": port,
                "properties": td,
                "addresses": [self.address],
            }
            service = ServiceInfo(**self.kwargs)
            log("ServiceInfo(%s)=%s", self.kwargs, service)
            self.service = service
        except Exception as e:
            log("zeroconf ServiceInfo", exc_info=True)
            log.error(" for port %i", port)
            log.estr(e)

    def start(self) -> None:
        try:
            self.zeroconf = Zeroconf(interfaces=[self.host])
        except OSError:
            log("start()", exc_info=True)
            log.error("Error: failed to create Zeroconf instance for address '%s'", self.host)
            return
        try:
            self.zeroconf.register_service(self.service)
        except Exception:
            log("start failed on %s", self.service, exc_info=True)
        else:
            self.registered = True

    def stop(self) -> None:
        log("ZeroConfPublishers.stop(): %s", self.service)
        if self.registered:
            self.zeroconf.unregister_service(self.service)
        self.zeroconf = None

    def update_txt(self, txt) -> None:
        if not hasattr(self.zeroconf, "update_service"):
            log("no update_service with zeroconf version %s", zeroconf_version)
            return
        props = txt_rec(txt)
        self.kwargs["properties"] = props
        si = ServiceInfo(**self.kwargs)
        try:
            self.zeroconf.update_service(si)
            self.service = si
        except KeyError as e:
            # probably a race condition with cleanup
            log("update_txt(%s)", txt, exc_info=True)
            log.warn("Warning: failed to update service")
            log.warn(" %s", e)
        except Exception:
            log.error("Error: failed to update service", exc_info=True)


def main():
    import random
    port = int(20000 * random.random()) + 10000
    # host = "127.0.0.1"
    host = "0.0.0.0"
    host_ports = [(host, port)]
    service_name = "test %s" % int(random.random() * 100000)
    from xpra.os_util import gi_import
    GLib = gi_import("GLib")
    publishers = []

    def add(service_type):
        publisher = ZeroconfPublishers(host_ports, service_name, service_type, {"somename": "somevalue"})
        GLib.idle_add(publisher.start)
        publishers.append(publisher)

        def update_rec():
            publisher.update_txt({"somename": "someothervalue"})

        GLib.timeout_add(10 * 1000, update_rec)

    add(XPRA_TCP_MDNS_TYPE)
    add(XPRA_UDP_MDNS_TYPE)
    loop = GLib.MainLoop()
    log.info("python-zeroconf version %s", zeroconf_version)
    try:
        loop.run()
    except KeyboardInterrupt:
        for publisher in publishers:
            publisher.stop()
        loop.quit()


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import struct

# Note: since encoding flags and compression flags are all mutually exclusive,
# (ie: only one encoder and at most one compressor can be used at a time)
# we could theoretically add many more values here,
# not necessarily limiting ourselves to the ones that land on a bit.

# packet encoding flags:
FLAGS_BENCODE = 0x0  # assume bencode if not other flag is set
FLAGS_RENCODE = 0x1
FLAGS_YAML = 0x4
FLAGS_RENCODEPLUS = 0x10

# these flags can actually be combined with the encoders above:
FLAGS_FLUSH = 0x8
FLAGS_CIPHER = 0x2

# compression flags are carried in the "level" field,
# the low bits contain the compression level, the high bits the compression algo:
LZ4_FLAG = 0x10
# LZO_FLAG        = 0x20
BROTLI_FLAG = 0x40
FLAGS_NOHEADER = 0x10000  # never encoded, so we can use a value bigger than a byte

_header_unpack_struct = struct.Struct(b'!cBBBL')
HEADER_SIZE = _header_unpack_struct.size
assert HEADER_SIZE == 8


def unpack_header(buf) -> tuple:
    return _header_unpack_struct.unpack_from(buf)


# 'P' + protocol-flags + compression_level + packet_index + data_size
_header_pack_struct = struct.Struct(b'!BBBBL')
assert ord("P") == 80


def pack_header(proto_flags: int, level: int, index: int, payload_size: int) -> bytes:
    return _header_pack_struct.pack(80, proto_flags, level, index, payload_size)


def find_xpra_header(data, index: int = 0, max_data_size: int = 2 ** 16) -> int:
    pos = data.find(b"P")
    while pos >= 0:
        if len(data) < pos + 8:
            # not enough data to try to parse this potential header
            return -1
        pchar, pflags, compress, packet_index, data_size = unpack_header(data[pos:pos + 8])
        if pchar == b"P" and packet_index == index and data_size < max_data_size:
            # validate flags:
            if compress == 0 or (compress & 0xf) > 0:
                # pylint: disable=import-outside-toplevel
                encoder_flag = pflags & (FLAGS_RENCODE | FLAGS_YAML | FLAGS_RENCODEPLUS)
                n_flags_set = sum(1 for flag in (FLAGS_RENCODE, FLAGS_YAML, FLAGS_RENCODEPLUS) if encoder_flag & flag)
                if encoder_flag == 0 or n_flags_set == 1:
                    return pos
        # skip to the next potential header:
        pos = data.find(b"P", pos + 1)
    return -1



# This file is part of Xpra.
# Copyright (C) 2024 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Callable

from xpra.os_util import gi_import
from xpra.scripts.pinentry import log
from xpra.util.thread import is_main_thread

Gtk = gi_import("Gtk")
GLib = gi_import("GLib")
Gio = gi_import("Gio")


def do_run_dialog(dialog: Gtk.Dialog) -> int:
    from xpra.platform.gui import force_focus
    try:
        force_focus()
        dialog.show()
        return dialog.run()
    finally:
        dialog.hide()
        dialog.close()


def dialog_run(run_fn: Callable) -> str | int:
    log("dialog_run(%s) is_main_thread=%s, main_level=%i", run_fn, is_main_thread(), Gtk.main_level())
    if is_main_thread() or Gtk.main_level() == 0:
        return run_fn()
    log("dialog_run(%s) main_depth=%s", run_fn, GLib.main_depth())
    # do a little dance if we're not running in the main thread:
    # block this thread and wait for the main thread to run the dialog
    from threading import Event
    e = Event()
    code = []

    def main_thread_run():
        log("main_thread_run() calling %s", run_fn)
        try:
            code.append(run_fn())
        finally:
            e.set()

    GLib.idle_add(main_thread_run)
    log("dialog_run(%s) waiting for main thread to run", run_fn)
    e.wait()
    log("dialog_run(%s) code=%s", run_fn, code)
    return code[0]


ICON_SIZES: dict[int, int] = {}
for size_name in ("MENU", "SMALL_TOOLBAR", "LARGE_TOOLBAR", "BUTTON", "DND", "DIALOG"):
    value = getattr(Gtk.IconSize, size_name, -1)
    if value >= 0:
        valid, width, height = Gtk.IconSize.lookup(value)
        if valid:
            ICON_SIZES[max(width, height)] = value


def nearest_icon_size(size: int) -> int:
    # try to find a size smaller or equal:
    best = 0
    for icon_size in ICON_SIZES.keys():
        if best < icon_size <= size:
            best = icon_size
    return best


def nearest_icon_enum(size: int) -> Gtk.IconSize:
    # try to find a size smaller or equal:
    best_size = nearest_icon_size(size)
    return ICON_SIZES.get(best_size, Gtk.IconSize.MENU)


def load_hb_image(icon_name: str, size=32) -> Gtk.Image | None:
    from xpra.gtk.pixbuf import get_icon_pixbuf
    pixbuf = get_icon_pixbuf(f"{icon_name}.png")
    if pixbuf:
        if pixbuf.get_width() != size or pixbuf.get_height() != size:
            GdkPixbuf = gi_import("GdkPixbuf")
            pixbuf = pixbuf.scale_simple(size, size, GdkPixbuf.InterpType.HYPER)
        return Gtk.Image.new_from_pixbuf(pixbuf)
    # try from the theme:
    icon = Gio.ThemedIcon(name=icon_name)
    if not icon:
        return None
    icon_size = nearest_icon_enum(size)
    return Gtk.Image.new_from_gicon(icon, icon_size)


def hb_button(tooltip: str, icon_name: str, callback: Callable, size=32) -> Gtk.Button:
    btn = Gtk.Button()
    image = load_hb_image(icon_name, size)
    if image:
        btn.add(image)
        btn.set_tooltip_text(tooltip)
        current_size = [size, size]

        # keep image square:
        def alloc(_btn, rect) -> None:
            if current_size != [rect.width, rect.height]:
                new_size = max(rect.width, rect.height)
                btn.set_size_request(new_size, new_size)
                scaled_image = load_hb_image(icon_name, new_size)
                if image:
                    btn.set_image(scaled_image)
                    current_size[0] = current_size[1] = new_size

        image.connect("size-allocate", alloc)

    def clicked(*_args) -> None:
        callback(btn)

    btn.connect("clicked", clicked)
    return btn



# -*- coding: utf-8 -*-
# This file is part of Xpra.
# Copyright (C) 2015 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# cityhash wrapper

#cython: boundscheck=False, wraparound=False

from libc.stdint cimport uint64_t, uint8_t

cdef extern from "city.h":
    uint64_t CityHash64(const char *buf, size_t len) nogil

cdef uint64_t cityhash64(uint8_t *data, size_t length) noexcept nogil:
    return CityHash64(<const char*> data, length)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Sequence

from xpra.util.str_fn import Ellipsizer
from xpra.os_util import gi_import
from xpra.log import Logger

log = Logger("keyboard")
verboselog = Logger("keyboard", "verbose")

# this allows platforms to inject keyname workarounds
# the key is a tuple (keyname, keyval, keycode)
# the value is the keyname override
KEY_TRANSLATIONS: dict[tuple[str, int, int], str] = {}


def get_default_keymap():
    Gdk = gi_import("Gdk")
    display = Gdk.Display.get_default()
    if not display:
        return Gdk.Keymap.get_default()
    return Gdk.Keymap.get_for_display(display)


def get_gtk_keymap(ignore_keys=("", "VoidSymbol", "0xffffff")) -> Sequence[tuple[int, str, int, int, int]]:
    """
        Augment the keymap we get from gtk.gdk.keymap_get_default()
        by adding the keyval_name.
        We can also ignore some keys
    """
    Gdk = gi_import("Gdk")
    keymap = get_default_keymap()
    if not keymap:
        return ()
    log("get_default_keymap(%s)=%s, direction=%s, bidirectional layouts: %s",
        ignore_keys, keymap, keymap.get_direction(), keymap.have_bidi_layouts())
    keycodes: list[tuple[int, str, int, int, int]] = []
    for i in range(0, 2 ** 8):
        entries = keymap.get_entries_for_keycode(i)
        if not entries:  # pragma: no cover
            continue
        found, keys, keyvals = entries
        if not found:
            verboselog("keycode %3i: ()", i)
            continue
        added = []
        for j, key in enumerate(keys):
            keyval = keyvals[j]
            keycode = key.keycode
            name = Gdk.keyval_name(keyval) or ""
            name = KEY_TRANSLATIONS.get((name, keyval, keycode), name)
            group = key.group or 0
            if name not in ignore_keys:
                kdef = keyval or 0, name or "", keycode or 0, group, key.level or 0
                keycodes.append(kdef)
                added.append(kdef)
        verboselog("keycode %3i: %s", i, added)
    log("get_gtk_keymap(%s)=%s (keymap=%s)", ignore_keys, Ellipsizer(keycodes), Ellipsizer(keymap))
    verboselog("get_gtk_keymap(%s)=%s (keymap=%s)", ignore_keys, keycodes, keymap)
    return tuple(keycodes)


def main() -> int:  # pragma: no cover
    # pylint: disable=import-outside-toplevel
    import sys
    from xpra.platform import program_context
    from xpra.log import enable_color, consume_verbose_argv
    with program_context("Keymap-Tool", "Keymap Information Tool"):
        enable_color()
        consume_verbose_argv(sys.argv, "keyboard")
        gtk_keymap = get_gtk_keymap()
        sizes = [16, 28, 8, 8, 8]

        def pkey(*entries) -> None:
            print(("".join([str(x).ljust(sizes[i]) for i, x in enumerate(entries)])).strip())

        pkey("keyval", "name", "keycode", "group", "level")
        for x in gtk_keymap:
            pkey(*x)
    return 0


if __name__ == "__main__":  # pragma: no cover
    main()



# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from libc.stdint cimport uintptr_t   # pylint: disable=syntax-error
from xpra.gtk.bindings.gobject cimport unwrap

from xpra.os_util import gi_import
from xpra.log import Logger

log = Logger("bindings", "gtk")

Gdk = gi_import("Gdk")


ctypedef float CGFloat
ctypedef int BOOL

cdef extern from "AppKit/AppKit.h":
    ctypedef struct NSColor:
        pass
    ctypedef struct NSWindow:
        pass
    ctypedef struct NSView:
        pass

cdef extern from "transparency_glue.h":
    #couldn't figure out how to get unions and cython+objc to play nice,
    #so we use a wrapper:
    void setAlphaValue(NSWindow *window, float alpha)
    void setOpaque(NSWindow *window, BOOL opaque)
    void setBackgroundColor(NSWindow *window, NSColor *color)
    void setClearBackgroundColor(NSWindow *window)
    void invalidateShadow(NSWindow *window)
    void setHasShadow(NSWindow *window, BOOL hasShadow)
    float getBackingScaleFactor(NSWindow *window)


cdef extern from "gtk-3.0/gdk/quartz/gdkquartz-cocoa-access.h":
    NSView *gdk_quartz_window_get_nsview(GdkWindow *window)
    NSWindow *gdk_quartz_window_get_nswindow(GdkWindow *window)


cdef extern from "gtk-3.0/gdk/gdk.h":
    ctypedef struct GdkWindow:
        pass

cdef GdkWindow *get_gdkwindow(pywindow):
    return <GdkWindow*>unwrap(pywindow, Gdk.Window)

cdef NSWindow *get_nswindow(pywindow):
    cdef GdkWindow *gdkwindow = get_gdkwindow(pywindow)
    assert gdkwindow
    cdef NSWindow *nswindow = gdk_quartz_window_get_nswindow(gdkwindow)
    return nswindow

cdef NSView *get_nsview(pywindow):
    cdef GdkWindow *gdkwindow = get_gdkwindow(pywindow)
    assert gdkwindow
    cdef NSView *nsview = gdk_quartz_window_get_nsview(gdkwindow)
    return nsview

def get_nsview_ptr(pywindow) -> int:
    cdef NSView *nsview = get_nsview(pywindow)
    return <uintptr_t> nsview

def enable_transparency(pywindow) -> None:
    cdef NSWindow *window = get_nswindow(pywindow)
    setClearBackgroundColor(window)
    setOpaque(window, 0)

def get_backing_scale_factor(pywindow) -> float:
    cdef NSWindow *window = get_nswindow(pywindow)
    return getBackingScaleFactor(window)



#!/usr/bin/env python3
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys

from xpra.os_util import gi_import
from xpra.platform import program_context
from xpra.platform.gui import force_focus
from xpra.gtk.window import add_close_accel
from xpra.gtk.pixbuf import get_icon_pixbuf

Gtk = gi_import("Gtk")
GLib = gi_import("GLib")


def change_callback(entry, window):
    print("text=%s" % entry.get_text())
    window.set_title(entry.get_text())


def make_window():
    window = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
    window.set_size_request(400, 100)
    window.set_position(Gtk.WindowPosition.CENTER)
    window.connect("delete_event", Gtk.main_quit)
    icon = get_icon_pixbuf("font.png")
    if icon:
        window.set_icon(icon)
    entry = Gtk.Entry()
    entry.set_max_length(50)
    entry.connect("changed", change_callback, window)
    title = "Hello"

    if len(sys.argv) > 1:
        title = sys.argv[1]
    entry.set_text(title)
    entry.show()
    window.add(entry)
    return window


def main():
    with program_context("window-title", "Window Title"):
        w = make_window()
        add_close_accel(w, Gtk.main_quit)
        from xpra.gtk.signals import quit_on_signals
        quit_on_signals("title test window")

        def show_with_focus():
            force_focus()
            w.show_all()
            w.present()

        GLib.idle_add(show_with_focus)
        Gtk.main()
        return 0


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import shlex
from subprocess import Popen
from collections.abc import Sequence

from xpra.util.objects import typedict
from xpra.util.str_fn import std, alnum, bytestostr
from xpra.util.env import envint, shellsub, first_time
from xpra.os_util import OSX, gi_import
from xpra.scripts.config import TRUE_OPTIONS
from xpra.util.child_reaper import getChildReaper
from xpra.auth.sys_auth_base import SysAuthenticator, log
from xpra.platform.features import EXECUTABLE_EXTENSION

GLib = gi_import("GLib")

TIMEOUT = envint("XPRA_EXEC_AUTH_TIMEOUT", 600)


def get_default_auth_dialog() -> str:
    if os.name == "posix":
        auth_dialog = "/usr/libexec/xpra/auth_dialog"
    else:
        from xpra.platform.paths import get_app_dir  # pylint: disable=import-outside-toplevel
        auth_dialog = os.path.join(get_app_dir(), "auth_dialog")
    if EXECUTABLE_EXTENSION:
        # ie: add ".exe" on MS Windows
        auth_dialog += "." + EXECUTABLE_EXTENSION
    log(f"auth_dialog={auth_dialog!r}")
    if not os.path.exists(auth_dialog) and first_time("auth-dialog-not-found"):
        log.warn(f"Warning: authentication dialog command {auth_dialog!r} does not exist")
    return auth_dialog


class Authenticator(SysAuthenticator):

    def __init__(self, **kwargs):
        log(f"exec.Authenticator({kwargs})")
        self.command = shlex.split(kwargs.pop("command", "${auth_dialog} ${info} ${timeout}"))
        self.require_challenge = kwargs.pop("require-challenge", "no").lower() in TRUE_OPTIONS
        self.timeout = kwargs.pop("timeout", TIMEOUT)
        self.timer = 0
        self.proc = None
        self.timeout_event = False
        if not self.command:
            self.command = [get_default_auth_dialog(), ]
        connection = kwargs.get("connection")
        if not connection:
            raise ValueError("connection object is missing")
        log(f"exec connection info: {connection}")
        self.connection_str = str(connection)
        super().__init__(**kwargs)

    def requires_challenge(self) -> bool:
        return self.require_challenge

    def get_challenge(self, digests: Sequence[str]) -> tuple[bytes, str]:
        assert self.require_challenge
        if "xor" not in digests:
            log.error("Error: kerberos authentication requires the 'xor' digest")
            return b"", ""
        return super().get_challenge(("xor", ))

    def validate_caps(self, caps: typedict) -> bool:
        if not self.require_challenge:
            return True
        return super().validate_caps(caps)

    def default_authenticate_check(self, caps: typedict) -> bool:
        info = f"Connection request from {self.connection_str}"
        subs = {
            "auth_dialog": get_default_auth_dialog(),
            "info": info,
            "timeout": self.timeout,
            "username": alnum(self.username),
            "prompt": std(self.prompt),
        }
        if self.require_challenge:
            subs["password"] = bytestostr(self.unxor_response(caps))
        cmd = tuple(shellsub(v, subs) for v in self.command)
        log(f"authenticate(..) shellsub({self.command}={cmd}")
        # [self.command, info, str(self.timeout)]
        try:
            with Popen(cmd) as proc:
                self.proc = proc
                log(f"authenticate(..) Popen({cmd})={proc}")
                # if required, make sure we kill the command when it times out:
                if self.timeout > 0:
                    self.timer = GLib.timeout_add(self.timeout * 1000, self.command_timedout)
                    if not OSX:
                        # python on macos may set a 0 returncode when we use poll()
                        # so we cannot use the ChildReaper on macos,
                        # and we can't cancel the timer
                        getChildReaper().add_process(proc, "exec auth", cmd, True, True, self.command_ended)
        except OSError as e:
            log(f"error running {cmd!r}", exc_info=True)
            log.error("Error: cannot run exec authentication module command")
            log.error(f" {cmd!r}")
            log.estr(e)
            return False
        v = proc.returncode
        log(f"authenticate(..) returncode({cmd})={v}")
        if self.timeout_event:
            return False
        return v == 0

    def command_ended(self, *args) -> None:
        t = self.timer
        log(f"exec auth.command_ended{args} timer={t}")
        if t:
            self.timer = 0
            GLib.source_remove(t)

    def command_timedout(self) -> None:
        proc = self.proc
        log(f"exec auth.command_timedout() proc={proc}")
        self.timeout_event = True
        self.timer = 0
        if proc:
            try:
                proc.terminate()
            except Exception:
                log(f"error trying to terminate exec auth process {proc}", exc_info=True)

    def __repr__(self):
        return "exec"



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.client.gui.window_border import WindowBorder


class AuthHandlersTest(unittest.TestCase):

	def test_toggle(self):
		b = WindowBorder(shown=True)
		b.toggle()
		assert b.shown is False

	def test_clone(self):
		b = WindowBorder(red=1, blue=0, green=1, alpha=0.5, size=10)
		b2 = b.clone()
		assert b.red==b2.red
		assert b.blue==b2.blue
		assert b.green==b2.green
		assert b.alpha==b2.alpha
		assert b.size==b2.size

	def test_repr(self):
		b = WindowBorder(red=0, blue=1, green=0.5)
		assert repr(b).find("00")>=0
		assert repr(b).find("FF")>=0


def main():
	unittest.main()


if __name__ == '__main__':
	main()



# This file is part of Xpra.
# Copyright (C) 2024 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from PyQt6.QtCore import Qt

key_names: dict[Qt.Key, str] = {}

for attr in dir(Qt.Key):
    if attr.startswith("Key_"):
        key_enum = getattr(Qt.Key, attr)
        key_names[key_enum] = attr[4:]



# This file is part of Xpra.
# Copyright (C) 2024 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.auth.sys_auth_base import SysAuthenticator, log
from xpra.util.objects import typedict


class Authenticator(SysAuthenticator):

    def __init__(self, **kwargs):
        log("http_header.Authenticator(%s)", kwargs)
        self.uid = -1
        self.gid = -1
        self.property = kwargs.pop("property", "X-Forwarded-Proto")
        self.value = str(kwargs.pop("value", "https"))
        connection = kwargs.get("connection", None)
        self.headers = getattr(connection, "options", {}).get("http-headers", {})
        log(f"http-headers({connection})={self.headers}")
        super().__init__(**kwargs)

    def get_uid(self) -> int:
        return self.uid

    def get_gid(self) -> int:
        return self.gid

    def requires_challenge(self) -> bool:
        return False

    def authenticate(self, caps: typedict) -> bool:  # pylint: disable=arguments-differ
        value = self.headers.get(self.property, "")
        log(f"{self.property!r}={value!r}, expected {self.value!r}")
        return value == self.value

    def __repr__(self):
        return "capability"



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2024 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import uuid
import signal
from time import monotonic
from collections.abc import Sequence
from typing import Any

from PyQt6.QtCore import QTimer, Qt
from PyQt6.QtNetwork import QTcpSocket
from PyQt6.QtWidgets import QApplication

from xpra import __version__
from xpra.exit_codes import ExitCode, ExitValue
from xpra.net.common import PacketType
from xpra.net.rencodeplus.rencodeplus import dumps, loads
from xpra.net.protocol.header import pack_header, unpack_header, FLAGS_RENCODEPLUS

from xpra.log import Logger

log = Logger("client")
netlog = Logger("network")


class Qt6Client:
    def __init__(self):
        self.windows: dict[int, Any] = {}
        self.socket = self.setup_socket()
        self.raw_packets = {}
        self.header = ()
        self.hello = {}
        self.focused = 0

    def setup_socket(self) -> QTcpSocket:
        socket = QTcpSocket()
        socket.disconnected.connect(self.socket_disconnected)
        socket.errorOccurred.connect(self.socket_error)
        socket.hostFound.connect(self.socket_hostfound)
        socket.stateChanged.connect(self.socket_statechanged)
        socket.readyRead.connect(self.socket_read)
        return socket

    def connect(self, host: str, port: int) -> None:
        log(f"connect({host}, {port})")
        self.socket.connectToHost(host, port)
        self.socket.connected.connect(self.socket_connected)

    def socket_error(self, *args) -> None:
        netlog(f"socket_error{args}")

    def socket_hostfound(self, *args) -> None:
        netlog(f"socket_hostfound{args}")

    def socket_statechanged(self, *args) -> None:
        netlog(f"socket_statechanged{args}")

    def socket_disconnected(self, *args) -> None:
        netlog(f"socket_disconnected{args}")
        self.quit(ExitCode.CONNECTION_LOST)

    def quit(self, exit_code: ExitValue):
        QApplication.exit(int(exit_code))

    def socket_connected(self, *args) -> None:
        netlog(f"socket_connected{args}")
        self.send_hello()

    def get_encodings(self) -> Sequence[str]:
        return "rgb", "png", "jpg", "webp"

    def send_hello(self) -> None:
        hello = {
            "version": __version__,
            "client_type": "qt6",
            "rencodeplus": True,
            "session-id": uuid.uuid4().hex,
            "windows": True,
            "keyboard": True,
            "mouse": True,
            "encodings": ("rgb32", "rgb24", "png", "jpg", "webp"),
            "network-state": False,  # tell older server that we don't have "ping"
        }
        self.send("hello", hello)

    def send(self, packet_type: str, *args) -> None:
        packet = (packet_type, *args)
        data = dumps(packet)
        header = pack_header(FLAGS_RENCODEPLUS, 0, 0, len(data))
        bin_packet = header+data
        self.socket.write(bin_packet)
        self.socket.flush()
        netlog(f"sent {packet_type!r}: {bin_packet!r}")

    def socket_read(self, *args) -> None:
        netlog(f"socket_read{args}")
        while avail := self.socket.bytesAvailable():
            need = 8 if not self.header else self.header[4]
            netlog(f" {need=}, available={avail}, header={self.header}")
            if avail < need:
                return
            if not self.header:
                assert need == 8 and avail >= 8
                header_bytes = self.socket.read(8)
                self.header = unpack_header(header_bytes)
                need = self.header[4]
                avail -= 8
            if avail < need:
                return
            packet_data = self.socket.read(need)
            netlog(f"read({need})={packet_data}")
            index = self.header[3]
            self.header = ()
            if index > 0:
                self.raw_packets[index] = packet_data
                continue
            packet = loads(packet_data)
            if self.raw_packets:
                packet = list(packet)
                for index, data in self.raw_packets.items():
                    packet[index] = data
                self.raw_packets = {}
            self.process_packet(packet)

    def process_packet(self, packet: PacketType) -> None:
        packet_type_fn_name = str(packet[0]).replace("-", "_")
        meth = getattr(self, f"_process_{packet_type_fn_name}", None)
        if not meth:
            netlog.warn(f"Warning: missing handler for {packet[0]!r}")
            netlog("packet=%r", packet)
            return
        meth(packet)

    def _process_hello(self, packet: PacketType) -> None:
        self.hello = packet[1]
        netlog.info("got hello from %s server" % self.hello.get("version", ""))

    def _process_setting_change(self, packet: PacketType) -> None:
        setting = packet[1]
        netlog.info(f"ignoring setting-change for {setting!r}")

    def _process_startup_complete(self, _packet: PacketType) -> None:
        netlog.info("client is connected")

    def _process_encodings(self, packet: PacketType) -> None:
        log(f"server encodings: {packet[1]}")

    def _process_new_window(self, packet: PacketType) -> None:
        self.new_window(packet)

    def _process_new_override_redirect(self, packet: PacketType) -> None:
        self.new_window(packet, True)

    def new_window(self, packet: PacketType, is_or=False) -> None:
        from xpra.client.qt6.window import ClientWindow
        wid, x, y, w, h = (int(item) for item in packet[1:6])
        metadata = packet[6]
        if is_or:
            metadata["override-redirect"] = is_or
        window = ClientWindow(self, wid, x, y, w, h, metadata)
        self.windows[wid] = window
        window.show()

    def _process_lost_window(self, packet: PacketType) -> None:
        wid = int(packet[1])
        window = self.windows.get(wid)
        if window:
            window.close()
            del self.windows[wid]

    def _process_raise_window(self, packet: PacketType) -> None:
        wid = int(packet[1])
        window = self.windows.get(wid)
        if window:
            window.raise_()

    def _process_draw(self, packet: PacketType) -> None:
        wid, x, y, width, height, coding, data, packet_sequence, rowstride = packet[1:10]
        window = self.windows.get(wid)
        now = monotonic()
        if window:
            message = ""
            window.draw(x, y, width, height, coding, data, rowstride)
            decode_time = int(1000 * (monotonic() - now))
        else:
            message = f"Warning: window {wid} not found"
            log.warn(message)
            decode_time = -1
        self.send("damage-sequence", packet_sequence, wid, width, height, decode_time, message)

    def _process_window_metadata(self, packet: PacketType) -> None:
        wid = packet[1]
        metadata = packet[2]
        log.info(f"window {wid}: {metadata}")

    def update_focus(self, wid=0) -> None:
        if self.focused == wid:
            return
        self.focused = wid

        def recheck_focus() -> None:
            if self.focused == wid:
                self.send("focus", wid, ())
        QTimer.singleShot(10, recheck_focus)

    def state_changed(self, state) -> None:
        log(f"state changed: {state}")
        if state == Qt.ApplicationState.ApplicationInactive:
            self.focused = 0
            self.send("focus", 0, ())


class XpraQt6Client(Qt6Client):
    """
    This class is just here to add the methods expected by the xpra.scripts.main script
    as these exist in the GTK3 client class.
    """

    def show_progress(self, pct: int, msg) -> None:
        log(f"show_progress({pct}, {msg})")

    def init(self, opts) -> None:
        """ we don't handle any options yet! """

    def init_ui(self, opts) -> None:
        """ we don't handle any options yet! """

    def cleanup(self) -> None:
        """ client classes must define this method """

    def make_protocol(self, conn):
        """
        Warnings:
            * this only works for plain `tcp` sockets
            * I couldn't figure out how to pass the existing socket file descriptor to QTcpSocket,
            so we re-open the connection instead...
        """
        from xpra.net.bytestreams import SocketConnection
        if not isinstance(conn, SocketConnection) or conn.socktype != "tcp":
            raise ValueError("only tcp socket connections are supported!")

        def connect():
            # self.socket.setSocketDescriptor(conn._socket.fileno())
            self.connect(*conn.remote)

        QTimer.singleShot(0, connect)

        class FakeProtocol:

            def start(self) -> None:
                log("FakeProtocol.start()")
                conn.close()
        return FakeProtocol()


def make_client() -> XpraQt6Client:
    app = QApplication([])
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    client = XpraQt6Client()
    app.applicationStateChanged.connect(client.state_changed)
    client.run = app.exec
    return client


def run_client(host: str, port: int) -> int:
    client = make_client()
    client.connect(host, port)
    return client.run()



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections import namedtuple
from typing import Any

from xpra.os_util import gi_import
from xpra.gtk.error import xlog
from xpra.x11.common import get_wm_name
from xpra.x11.desktop.model_base import DesktopModelBase
from xpra.util.rectangle import rectangle
from xpra.log import Logger

log = Logger("server")

GObject = gi_import("GObject")

MIN_SIZE = 640, 350
MAX_SIZE = 8192, 8192

MonitorDamageNotify = namedtuple("MonitorDamageNotify", ("x", "y", "width", "height"))


class MonitorDesktopModel(DesktopModelBase):
    """
    A desktop model representing a single monitor
    """
    __gsignals__ = dict(DesktopModelBase.__common_gsignals__)

    # bump the number of receivers,
    # because we add all the monitor models as receivers for the root window:
    MAX_RECEIVERS = 20

    def __repr__(self):
        return f"MonitorDesktopModel({self.name} : {self.monitor_geometry})"

    def __init__(self, monitor: dict[str, Any]):
        super().__init__()
        self.init(monitor)

    def init(self, monitor: dict[str, Any]) -> None:
        self.name = monitor.get("name", "")
        self.resize_delta = 0, 0
        x = monitor.get("x", 0)
        y = monitor.get("y", 0)
        width = monitor.get("width", 0)
        height = monitor.get("height", 0)
        self.monitor_geometry = (x, y, width, height)
        self._updateprop("size-hints",
                         {
                             "minimum-size": MIN_SIZE,
                             "maximum-size": MAX_SIZE,
                         })

    def get_title(self) -> str:
        with xlog:
            title = get_wm_name()  # pylint: disable=assignment-from-none
        if self.name:
            if not title:
                return self.name
            title += f" on {self.name}"
        return title

    def get_geometry(self) -> tuple[int, int, int, int]:
        return self.monitor_geometry

    def get_dimensions(self) -> tuple[int, int]:
        return self.monitor_geometry[2], self.monitor_geometry[3]

    def get_definition(self) -> dict[str, Any]:
        x, y, width, height = self.monitor_geometry
        return {
            "geometry": self.monitor_geometry,
            "x": x,
            "y": y,
            "width": width,
            "height": height,
            "name": self.name,
        }

    def do_x11_damage_event(self, event) -> None:
        # ie: <X11:DamageNotify {'send_event': '0', 'serial': '0x4da', 'delivered_to': '0x56e', 'window': '0x56e',
        #                       'damage': '2097157', 'x': '313', 'y': '174', 'width': '6', 'height': '13'}>)
        damaged_area = rectangle(event.x, event.y, event.width, event.height)
        x, y, width, height = self.monitor_geometry
        monitor_damaged_area = damaged_area.intersection(x, y, width, height)
        if monitor_damaged_area:
            # use an event relative to this monitor's coordinates:
            mod_event = MonitorDamageNotify(monitor_damaged_area.x - x, monitor_damaged_area.y - y,
                                            monitor_damaged_area.width, monitor_damaged_area.height)
            self.emit("client-contents-changed", mod_event)

    def get_image(self, x: int, y: int, width: int, height: int):
        # adjust the coordinates with the monitor's position:
        mx, my = self.monitor_geometry[:2]
        image = super().get_image(mx + x, my + y, width, height)
        if image:
            image.set_target_x(x)
            image.set_target_y(y)
        return image

    def do_resize(self) -> None:
        self.resize_timer = 0
        x, y, saved_width, saved_height = self.monitor_geometry
        width, height = self.resize_value
        self.monitor_geometry = (x, y, width, height)
        self.resize_delta = width - saved_width, height - saved_height
        self.emit("resized")


GObject.type_register(MonitorDesktopModel)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.os_util import OSX, POSIX
from unit.server.mixins.server_mixins_option_test_util import ServerMixinsOptionTestUtil


class StartDesktopOptionTest(ServerMixinsOptionTestUtil):

    def test_start_all(self):
        self._test_all("start-desktop")


def main():
    if POSIX and not OSX:
        unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


#new recommended way of using the glib main loop:
from dbus.mainloop.glib import DBusGMainLoop
DBusGMainLoop(set_as_default=True)
import dbus   # pylint: disable=wrong-import-position


#NAME = "org.freedesktop.ScreenSaver"
#PATH = "/org/freedesktop/ScreenSaver"
NAME = "org.gnome.ScreenSaver"
PATH = "/org/gnome/ScreenSaver"

def main():
    from xpra.platform import program_context
    with program_context("ScreenSaver-listener", "ScreenSaver listener"):
        dbus_session = dbus.SessionBus()
        def active_changed(active):
            print("screensaver active status changed: %s" % (active, ))
        dbus_session.add_signal_receiver(active_changed, "ActiveChanged", NAME, path=PATH)
        from gi.repository import GLib  # @UnresolvedImport
        loop = GLib.MainLoop()
        loop.run()


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

display = True
windows = True
cursors = True
gstreamer = True
x11 = True
webcam = True
audio = True
clipboard = True
notifications = True
dbus = True
mmap = True
ssl = True
ssh = True
logging = True
tray = True
network_state = True
network_listener = True
encoding = True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
import unittest
import binascii

from xpra.util.str_fn import Ellipsizer
from xpra.log import Logger, consume_verbose_argv

log = Logger("brotli")


def b(s) -> bytes:
    return binascii.unhexlify(s.replace(" ", ""))


INVALID_INPUTS = (None, True, 1, 1.2, [1, 2], (1, 2), object())


# pylint: disable=import-outside-toplevel


class TestBrotli(unittest.TestCase):

    def test_libversions(self) -> None:
        from xpra.net.brotli import decompressor, compressor  # @UnresolvedImport
        for m in (decompressor, compressor):
            v = m.get_version()
            log("%s.get_version()=%s", m, v)
            assert v and len(v) == 3
            assert v[0] >= 1

    def td(self, v, match_value=None, maxsize=512 * 1024):
        log("tc%s", (Ellipsizer(v), Ellipsizer(match_value), maxsize))
        from xpra.net.brotli.decompressor import decompress  # @UnresolvedImport
        value = decompress(v, maxsize)
        if match_value is not None:
            assert value == match_value, f"expected {match_value!r} but got {value!r}"
        return value

    def fd(self, v, match_value=None, maxsize=512 * 1024):
        try:
            self.td(v, match_value, maxsize)
        except Exception:
            pass
        else:
            raise ValueError(f"decompression should have failed for {v!r}")

    def tc(self, v, match_value=None, level=2, maxsize=512 * 1024):
        log("tc%s", (Ellipsizer(v), Ellipsizer(match_value), level, maxsize))
        from xpra.net.brotli.compressor import compress  # @UnresolvedImport
        value = compress(v, level)
        if match_value is not None:
            assert value == match_value, f"expected {match_value!r} but got {value!r}"
        return value

    def fc(self, v, match_value=None, maxsize=512 * 1024):
        try:
            self.tc(v, match_value, maxsize)
        except Exception:
            pass
        else:
            raise ValueError(f"compression should have failed for {v!r}")

    def test_decompressinvalidinput(self):
        for v in INVALID_INPUTS:
            self.fd(v)
        self.fd(b"hello")

    def test_decompressinputtoosmall(self):
        self.fd(b"hello")

    def test_decompressvalid(self):
        self.td(b("2110000468656c6c6f03"), b"hello")

    def test_limit(self):
        br = b("59 ff ff 8f 5f 02 26 1e 0b 04 72 ef 1f 00")
        self.td(br, b"0" * 1024 * 1024, 1024 * 1024)
        self.fd(br, b"0" * 1024 * 1024, 1024 * 1024 - 1)

    def test_compress(self):
        for l in range(2, 11):
            self.tc(b"hello", b'\x0b\x02\x80hello\x03', l)

    def test_compressinvalidinput(self):
        for v in INVALID_INPUTS:
            self.fc(v)

    def test_roundtrip(self):
        TEST_INPUT = [b"hello", b"*" * 1024, b"+" * 64 * 1024]

        #find some real "text" files:

        def addf(path) -> None:
            if path and os.path.exists(path):
                with open(path, "rb") as f:
                    TEST_INPUT.append(f.read())

        if __file__:
            addf(__file__)
            path = os.path.abspath(os.path.dirname(__file__))
            while path:
                addf(os.path.join(path, "COPYING"))
                addf(os.path.join(path, "README.md"))
                parent_path = os.path.abspath(os.path.join(path, os.pardir))
                if parent_path == path or not parent_path:
                    break
                path = parent_path
        for l in range(2, 11):
            for v in TEST_INPUT:
                c = self.tc(v, level=l)
                self.td(c, v)


def main() -> None:
    consume_verbose_argv(sys.argv, "brotli")
    try:
        from xpra.net.brotli import decompressor, compressor
        assert decompressor and compressor
    except ImportError as e:
        print(f"brotli test skipped: {e}")
    else:
        unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import json
from collections.abc import Iterable, Sequence

from xpra.platform.keyboard_base import KeyboardBase
from xpra.keyboard.mask import MODIFIER_MAP
from xpra.log import Logger
from xpra.util.env import first_time
from xpra.util.system import is_Wayland, is_X11
from xpra.util.str_fn import bytestostr


log = Logger("keyboard", "posix")


def get_locale_status() -> dict[str, str]:
    # parse the output into a dictionary:
    # $ localectl status
    # System Locale: LANG=en_GB.UTF-8
    # VC Keymap: gb
    # X11 Layout: gb
    from subprocess import getoutput  # pylint: disable=import-outside-toplevel
    out = getoutput("localectl status")
    if not out:
        return {}
    locale = {}
    for line in bytestostr(out).splitlines():
        parts = line.lstrip(" ").split(": ")
        if len(parts) == 2:
            locale[parts[0]] = parts[1]
    log("locale(%s)=%s", out, locale)
    return locale


def get_xkb_rules_names_property() -> Sequence[str]:
    # parses the "_XKB_RULES_NAMES" X11 property
    if not is_X11():
        return ()
    xkb_rules_names: list[str] = []
    # pylint: disable=import-outside-toplevel
    from xpra.gtk.error import xlog
    from xpra.x11.common import get_X11_root_property
    with xlog:
        prop = get_X11_root_property("_XKB_RULES_NAMES", "STRING")
        log("get_xkb_rules_names_property() _XKB_RULES_NAMES=%s", prop)
        # ie: 'evdev\x00pc104\x00gb,us\x00,\x00\x00'
        if prop:
            xkb_rules_names = prop.decode("latin1").split("\0")
        # ie: ['evdev', 'pc104', 'gb,us', ',', '', '']
    log("get_xkb_rules_names_property()=%s", xkb_rules_names)
    return tuple(xkb_rules_names)


def get_all_x11_layouts() -> dict[str, str]:
    repository = "/usr/share/X11/xkb/rules/base.xml"
    x11_layouts: dict[str, str] = {}
    if os.path.exists(repository):
        try:
            import lxml.etree  # pylint: disable=import-outside-toplevel
        except ImportError:
            log("cannot parse xml", exc_info=True)
        else:
            try:
                with open(repository, encoding="latin1") as f:
                    tree = lxml.etree.parse(f)  # pylint: disable=c-extension-no-member @UndefinedVariable
            except Exception:
                log.error(f"Error parsing {repository}", exc_info=True)
            else:
                for layout in list(tree.xpath("//layout")):
                    layout_str = str(layout.xpath("./configItem/name")[0].text)
                    x11_layouts[layout] = layout_str
                    # for variant in layout.xpath("./variantlist/variant/configItem/name"):
                    #    variant_name = variant.text
                return x11_layouts
    from subprocess import Popen, PIPE  # pylint: disable=import-outside-toplevel
    try:
        proc = Popen(["localectl", "list-x11-keymap-layouts"], stdout=PIPE, stderr=PIPE)
        out = proc.communicate()[0]
        log("get_all_x11_layouts() proc=%s", proc)
        log("get_all_x11_layouts() returncode=%s", proc.returncode)
        if proc.wait() == 0:
            for line in out.splitlines():
                layout = line.decode().split("/")[-1]
                if layout:
                    x11_layouts[layout] = layout
            return x11_layouts
    except OSError:
        log("get_all_x11_layouts()", exc_info=True)
    return {"us": "English"}


class Keyboard(KeyboardBase):

    def __init__(self):
        super().__init__()
        if is_X11():
            try:
                # pylint: disable=import-outside-toplevel
                from xpra.x11.bindings.keyboard import X11KeyboardBindings
                self.keyboard_bindings = X11KeyboardBindings()
            except Exception as e:
                log("keyboard bindings", exc_info=True)
                from xpra.gtk.util import ds_inited
                if not ds_inited():
                    log.error("Error: failed to load the X11 keyboard bindings")
                    log.error(" %s", str(e) or type(e))
                    log.error(" keyboard mapping may be incomplete")

    def init_vars(self) -> None:
        super().init_vars()
        self.keymap_modifiers = None
        self.keyboard_bindings = None
        self.__input_sources: dict[str, int] = {}
        try:
            from xpra.dbus.helper import DBusHelper
            self.__dbus_helper = DBusHelper()
            # we really want to catch `dbus.exceptions.DBusException` here instead:
        except Exception as e:
            self.__dbus_helper = None
            log.info(f"gnome input sources requires dbus: {e}")
        else:
            self._dbus_gnome_shell_ism("List", [], self._store_input_sources)

    def _store_input_sources(self, input_sources: Iterable) -> None:
        log("_store_input_sources(%s)", input_sources)
        for layout_info in input_sources.values():
            index = int(layout_info["index"])
            layout_variant = str(layout_info["id"])
            layout = layout_variant.split("+", 1)[0]
            self.__input_sources[layout] = index

    def _dbus_gnome_shell_ism(self, method, args, callback=None) -> None:
        from xpra.dbus.helper import dbus_to_native

        def ok_cb(success, res) -> None:
            try:
                if not dbus_to_native(success):
                    log("_dbus_gnome_shell_ism(method=%s, args=%s): %s", method, args, success)
                    return
                if callback is not None:
                    callback(json.loads(dbus_to_native(res)))
            except Exception:
                log("_dbus_gnome_shell_ism(method=%s, args=%s)", method, args, exc_info=True)

        def err_cb(msg) -> None:
            log("_dbus_gnome_shell_ism(method=%s, args=%s): %s", method, args, msg)
            if method == "List":
                self._dbus_gnome_shell_eval_ism(".inputSources", self._store_input_sources)
            elif method == "Activate":
                try:
                    index = dbus_to_native(args[0])
                except Exception:
                    log("_dbus_gnome_shell_ism(method=%s, args=%s): the Eval fallback", method, args, exc_info=True)
                else:
                    self._dbus_gnome_shell_eval_ism(f".inputSources[{index}].activate()")

        log("_dbus_gnome_shell_ism(method=%s, args=%s): before call_function", method, args)
        self.__dbus_helper.call_function(
            "org.gnome.Shell",
            "/org/xpra/InputSourceManager",
            "xpra_org.InputSourceManager",
            method,
            args,
            ok_cb,
            err_cb,
        )

    def _dbus_gnome_shell_eval_ism(self, cmd, callback=None) -> None:
        from xpra.dbus.helper import dbus_to_native, native_to_dbus
        ism = "imports.ui.status.keyboard.getInputSourceManager()"

        def ok_cb(success, res) -> None:
            try:
                if not dbus_to_native(success):
                    log("_dbus_gnome_shell_eval_ism(%s): %s", cmd, success)
                    return
                if callback is not None:
                    callback(json.loads(dbus_to_native(res)))
            except Exception:
                log("_dbus_gnome_shell_eval_ism(%s)", cmd, exc_info=True)

        def err_cb(msg) -> None:
            log("_dbus_gnome_shell_eval_ism(%s): %s", cmd, msg)

        self.__dbus_helper.call_function(
            "org.gnome.Shell",
            "/org/gnome/Shell",
            "org.gnome.Shell",
            "Eval",
            [native_to_dbus(ism + cmd)],
            ok_cb,
            err_cb,
        )

    def set_platform_layout(self, layout: str) -> None:
        index = self.__input_sources.get(layout)
        log("set_platform_layout(%s): index=%s", layout, index)
        if index is None:
            log(f"asked layout ({layout}) has no corresponding registered input source")
            return
        if self.__dbus_helper:
            from xpra.dbus.helper import native_to_dbus
            self._dbus_gnome_shell_ism("Activate", [native_to_dbus(index)])

    def __repr__(self):
        return "posix.Keyboard"

    def get_keymap_modifiers(self) -> tuple[dict, list[str], list[str]]:
        if self.keymap_modifiers is None:
            self.keymap_modifiers = self.do_get_keymap_modifiers()
        return self.keymap_modifiers

    def do_get_keymap_modifiers(self) -> tuple[dict, list[str], list[str]]:
        if not self.keyboard_bindings:
            if is_Wayland():
                if first_time("wayland-keymap"):
                    log.warn("Warning: incomplete keymap support under Wayland")
                return {}, [], ["mod2", ]
            return {}, [], []
        from xpra.gtk.error import xlog
        with xlog:
            mod_mappings = self.keyboard_bindings.get_modifier_mappings()
            if mod_mappings:
                # ie: {"shift" : ["Shift_L", "Shift_R"], "mod1" : "Meta_L", ...]}
                log("modifier mappings=%s", mod_mappings)
                meanings: dict[str, str] = {}
                for modifier, keys in mod_mappings.items():
                    for _, keyname in keys:
                        meanings[keyname] = modifier
                # probably a GTK bug? but easier to put here
                mod_missing: list[str] = []
                numlock_mod = meanings.get("Num_Lock", "")
                if numlock_mod:
                    mod_missing.append(numlock_mod)
                return meanings, [], mod_missing
        return {}, [], []

    # noinspection PyUnreachableCode
    def get_x11_keymap(self) -> dict[int, list[str]]:
        if not self.keyboard_bindings:
            return {}
        from xpra.gtk.error import xlog
        with xlog:
            return self.keyboard_bindings.get_keycode_mappings()
        return {}

    def get_keymap_spec(self) -> dict[str, str]:
        log("get_keymap_spec() keyboard_bindings=%s", self.keyboard_bindings)
        if is_Wayland() or not self.keyboard_bindings:
            locale = get_locale_status()
            query_struct = {}
            if locale:
                layout = locale.get("X11 Layout")
                if layout:
                    query_struct["layout"] = layout
            log("query_struct(%s)=%s", locale, query_struct)
            return query_struct
        from xpra.gtk.error import xsync
        with xsync:
            query_struct = self.keyboard_bindings.getXkbProperties()
        log("get_keymap_spec()=%r", query_struct)
        return query_struct

    def get_all_x11_layouts(self):
        return get_all_x11_layouts()

    def get_layout_spec(self) -> tuple[str, str, Sequence[str], str, Sequence[str], str]:
        model = ""
        layout = ""
        layouts = ()
        variant = ""
        options = ""
        if self.keyboard_bindings:
            from xpra.gtk.error import xsync
            with xsync:
                props = self.keyboard_bindings.getXkbProperties()
            model = props.get("model", "")
            v = props.get("layout")
            variant = props.get("variant", "")
            options = props.get("options", "")
        else:
            locale = get_locale_status()
            v = locale.get("X11 Layout", "")
        if not v:
            # fallback:
            props = get_xkb_rules_names_property()
            # ie: ['evdev', 'pc104', 'gb,us', ',', '', '']
            if props and len(props) >= 3:
                v = props[2]
        if v:
            layouts = v.split(",")
            layout = v
        return model, layout, layouts, variant, (), options

    def get_keyboard_repeat(self) -> tuple[int, int] | None:
        if self.keyboard_bindings:
            try:
                v = self.keyboard_bindings.get_key_repeat_rate()
                if v:
                    assert len(v) == 2
                    log("get_keyboard_repeat()=%s", v)
                    return v[0], v[1]
            except Exception as e:
                log.error("Error: failed to get keyboard repeat rate:")
                log.estr(e)
        return None

    def update_modifier_map(self, mod_meanings) -> None:
        try:
            # pylint: disable=import-outside-toplevel
            from xpra.x11.gtk.keys import grok_modifier_map
            self.modifier_map = grok_modifier_map(mod_meanings)
        except ImportError:
            self.modifier_map = MODIFIER_MAP
        # force re-query on next call:
        self.keymap_modifiers = None
        log(f"update_modifier_map({mod_meanings}) modifier_map={self.modifier_map}")



# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


import sys
from xpra.net.vsock import connect_vsocket, CID_ANY, PORT_ANY       #@UnresolvedImport

from xpra.log import Logger
log = Logger("network")


def main(args):
    cid = CID_ANY
    port = PORT_ANY
    if len(args)>0:
        cid = int(args[0])
    if len(args)>1:
        port = int(args[1])

    vsock = connect_vsocket(cid=cid, port=port)
    log("vsock=%s", vsock)
    vsock.send(" "*1024*1024*1024)
    data = vsock.recv(1024)
    log("recv()=%s", data)

if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from time import monotonic

from xpra.util.env import numpy_import_context
from xpra.codecs.image import ImageWrapper
from xpra.log import Logger

log = Logger("cuda")

# this module is only imported by nvfbc.capture_*
# and it may wait for CUDA initialization to complete here:
with numpy_import_context("CUDA image wrapper import", True):
    from numpy import byte  # @UnresolvedImport
    from pycuda.driver import (
        pagelocked_empty,
        memcpy_dtoh_async,
        LogicError,
    )


class CUDAImageWrapper(ImageWrapper):

    def __init__(self, *args):
        super().__init__(*args)
        self.stream = None
        self.cuda_device_buffer = None
        self.cuda_context = None
        self.buffer_size = 0

    def wait_for_stream(self):
        s = self.stream
        if s and not s.is_done():
            self.stream.synchronize()

    def may_download(self):
        ctx = self.cuda_context
        if self.pixels is not None or not ctx or self.freed:
            return
        assert self.cuda_device_buffer, "bug: no device buffer"
        start = monotonic()
        ctx.push()
        host_buffer = pagelocked_empty(self.buffer_size, dtype=byte)  # pylint: disable=no-member
        memcpy_dtoh_async(host_buffer, self.cuda_device_buffer, self.stream)  # pylint: disable=no-member
        self.wait_for_stream()
        self.pixels = host_buffer.tobytes()
        elapsed = monotonic() - start
        if elapsed > 0:
            mbs = self.buffer_size / elapsed / 1024 / 1024
        else:
            mbs = 9999
        log("may_download() from %#x to %s, size=%s, elapsed=%ims - %iMB/s",
            int(self.cuda_device_buffer), host_buffer, self.buffer_size,
            int(1000 * elapsed), mbs)
        self.free_cuda()
        ctx.pop()

    def freeze(self) -> bool:
        # this image is already a copy when we get it
        return True

    def get_gpu_buffer(self):
        self.wait_for_stream()
        return self.cuda_device_buffer

    def has_pixels(self) -> bool:
        return self.pixels is not None

    def get_pixels(self):
        self.may_download()
        return super().get_pixels()

    def clone_pixel_data(self) -> None:
        self.may_download()
        super().clone_pixel_data()

    def get_sub_image(self, x, y, w, h):
        self.may_download()
        return super().get_sub_image(x, y, w, h)

    def free_cuda_device_buffer(self) -> None:
        cdb = self.cuda_device_buffer
        if not cdb:
            return
        log("%s.free_cuda() cuda_device_buffer=%#x", self, int(cdb or 0))
        self.cuda_device_buffer = None
        cdb.free()

    def free_cuda(self) -> None:
        self.free_cuda_device_buffer()
        self.stream = None
        self.cuda_context = None
        self.buffer_size = 0

    def free(self):
        self.free_cuda()
        return super().free()

    def clean(self) -> None:
        try:
            self.wait_for_stream()
        except LogicError:  # pylint: disable=no-member @UndefinedVariable
            log("%s.clean()", self, exc_info=True)
        self.free()
